import re
import sys
import json
import types
import pprint
from pathlib import PurePath
from itertools import chain, zip_longest
from collections import Counter, defaultdict
import idlib
import htmlfn as hfn
import augpathlib as aug
from hyputils import hypothesis as hyp
from pyontutils.core import OntGraph, OntRes, OntResIri, OntResPath
from pyontutils.utils import UTCNOWISO, anyMembers, Async, deferred
from pyontutils.namespaces import isAbout, ilxtr, TEMP
from sparcur import sheets
from sparcur.core import (OntId,
                          OntTerm,
                          get_all_errors,
                          adops,
                          JApplyRecursive,
                          get_nested_by_key,
                          get_nested_by_type,
                          )
from sparcur.utils import want_prefixes, log as _log, PennsieveId
from sparcur.utils import expand_label_curie, is_list_or_tuple
from sparcur.paths import Path
from sparcur.config import auth

log = _log.getChild('reports')


def normalize_exact(e, *, lower=False):
    if e is not None:
        e = e.replace('\n', ' ')  # sigh
        e = e.replace('\t', ' ')
        e = e.replace('\xA0', ' ')  # nbsp
        e = e.strip().rstrip()
        if lower:
            e = e.lower()
        return e

    return ''


def normtext(t, *, lower=False):
    if 'https://' in t:
        return ''
    else:
        t = t.strip().rstrip()
        if lower:
            t = t.lower()

        return t


class SparqlQueries:
    """ Creates SPARQL query templates. """

    # XXX These queries now live in docs/queries.org and are maintained there
    # This class is deprecated and will be removed at some point in the future
    # it should not be updated

    def __init__(self, nsm=None):
        from rdflib.plugins import sparql
        self.sparql = sparql
        self.nsm = nsm if nsm else OntGraph().namespace_manager
        self.prefixes = dict(self.nsm)
        self.prefixes.update(
            {
                'protcur': 'https://uilx.org/tgbugs/u/protcur/',
                'ilxtr': str(ilxtr),  # apparently we cleaned up the dangling usages
             },)

    def dataset_about(self):
        # FIXME this will return any resource matching isAbout:
        query = """
            SELECT ?dataset
            WHERE {
                ?dataset rdf:type sparc:Dataset .
                ?dataset isAbout: ?about .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def dataset_subjects(self) -> str:
        """ Get all subject groups and dataset associated with subject input.

        :returns: list of tuples containing: subject, subjects group, and subjects dataset.
        """
        query = """
            SELECT ?dataset ?subj
            WHERE {
                ?startsubj TEMP:hasDerivedInformationAsParticipant ?dataset .
                ?subj  TEMP:hasDerivedInformationAsParticipant ?dataset .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def dataset_groups(self) -> str:
        """ Get all subject groups and dataset associated with subject input.

        :returns: list of tuples containing: subject, subjects group, and subjects dataset.
        """
        query = """
            SELECT ?dataset ?group ?subj
            WHERE {
                ?startsubj TEMP:hasDerivedInformationAsParticipant ?dataset .
                ?subj  TEMP:hasDerivedInformationAsParticipant ?dataset .
                ?subj  TEMP:hasAssignedGroup ?group .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def dataset_bundle(self) -> str:
        """ Get all related datasets of subject.

        :returns: list of tuples containing: subject & subjects shared dataset.
        """
        query = """
            SELECT ?dataset
            WHERE {
                ?startdataset TEMP:collectionTitle ?string .
                ?dataset  TEMP:collectionTitle ?string .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def dataset_subject_species(self):
        # FIXME how to correctly init bindings to multiple values ...
        query = """
            SELECT DISTINCT ?dataset
            WHERE {
                ?dataset TEMP:isAboutParticipant ?subject .
                ?subject sparc:animalSubjectIsOfSpecies ?species .
                FILTER ( CONTAINS(str(?species), "human")
                      || CONTAINS(str(?species), "homo sapiens")
                      || ?species = NCBITaxon:9606 )
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def award_affiliations(self):
        query = """
            SELECT DISTINCT ?award ?affiliation
            WHERE {
                ?dataset TEMP:hasAwardNumber ?award .
                ?contributor TEMP:contributorTo ?dataset .
                ?contributor TEMP:hasAffiliation ?affiliation .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def dataset_milestone_completion_date(self):
        query = """
            SELECT DISTINCT ?dataset ?date
            WHERE {
                ?dataset TEMP:milestoneCompletionDate ?date .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def protocol_techniques(self):
        query = """
            SELECT DISTINCT ?protocol ?technique
            WHERE {
                ?protocol rdf:type sparc:Protocol .
                ?protocol TEMP:protocolEmploysTechnique ?technique .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def protocol_aspects(self):
        query = """
            SELECT DISTINCT ?protocol ?aspect
            WHERE {
                ?protocol rdf:type sparc:Protocol .
                ?protocol TEMP:protocolInvolvesAspect ?ast .
                ?ast rdf:type protcur:aspect .
                ?ast TEMP:hasValue ?aspect .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def protocol_inputs(self):
        query = """
            SELECT DISTINCT ?protocol ?ast_in ?input
            WHERE {
                ?protocol rdf:type sparc:Protocol .
                ?protocol TEMP:protocolInvolvesInput ?ast_in .
                ?ast_in rdf:type protcur:input .
                ?ast_in TEMP:hasValue ?input .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def protocol_species_dose(self):
        query = """
SELECT DISTINCT
?dataset
?protocol

?label_drug
?value_lt
WHERE {
    VALUES ?t {protcur:invariant protcur:parameter} .
    ?ast_inv a ?t .
    ?ast_inv TEMP:hasValue ?bnode .
    ?bnode ilxtr:jsonRecordType <https://uilx.org/tgbugs/u/sparcur-protcur-json-ld/quantity> .
    ?bnode rdf:value ?value_lt .
    ?bnode TEMP:hasUnit <https://uilx.org/tgbugs/u/sparcur-protcur-json-ld/milligram%20%2F%20kilogram> .
    FILTER (?value_lt < ?limit)

    ?ast_drug a protcur:input .
    ?ast_drug TEMP:hasValue ?drug .
                            ?drug rdfs:label ?label_drug .
    ?ast_drug TEMP:protcurChildren+ ?ast_child .
    ?ast_child TEMP:hasValue ?bnode .

    ?protocol a sparc:Protocol .
    ?protocol TEMP:protocolInvolvesInput ?ast_drug .

    ?protocol TEMP:protocolInvolvesInput ?ast_in_sp .
    ?ast_in_sp rdf:type protcur:input .
    ?ast_in_sp TEMP:hasValue ?species .

    OPTIONAL { ?dataset TEMP:hasProtocol ?protocol } .

} ORDER BY ?label_input ?value_lt
"""
        return self.sparql.prepareQuery(query, initNs=self.prefixes)


class TtlFile:

    def __init__(self, ontres, ontres_compare_to=None):
        import rdflib
        from pyontutils import namespaces as ns
        self.rdflib = rdflib
        self.ns = ns

        self.ontres = ontres

        if ontres_compare_to is None:
            self.ontres_compare_to = ontres
        else:
            self.ontres_compare_to = ontres_compare_to

        self.graph = self.ontres.graph
        self.graph_compare_to = self.ontres_compare_to.graph

        self.queries = SparqlQueries(nsm=self.graph.namespace_manager)

    def _to_human(self, row):
        return [self.graph.namespace_manager.qname(cell)
                if isinstance(cell, self.rdflib.URIRef) else
                cell.toPython() for cell in row]

    def changes(self):
        added, removed, changed = self.graph.subjectsChanged(self.graph_compare_to)
        return added, removed, changed

    def milestones(self, ext=None):
        query = self.queries.dataset_milestone_completion_date()
        res = list(self.graph.query(query))
        rows = sorted(self._to_human(row) for row in res)
        header = [['dataset', 'milestone comp date']]
        return header + rows

    def mis(self, ext=None):
        from pyontutils.namespaces import OntCuries
        from pyontutils.config import auth as pauth
        rdflib = self.rdflib
        ns = self.ns
        ontres = self.ontres

        if False:
            from ttlser import CustomTurtleSerializer
            CustomTurtleSerializer.addTopClasses(
                ns.ilxtr.MISType,
                ns.ilxtr.MISPredicate,
            )

        preds = ontres.graph.predicates()
        types = (o for s, o in ontres.graph[:ns.rdf.type:])
        g = OntGraph(namespace_manager=ontres.graph.namespace_manager)
        OntCuries.populate(g.namespace_manager)
        for things, type_ in ((preds, ns.ilxtr.MISPredicate), (types, ns.ilxtr.MISType)):
            counts = Counter(things)

            [(g.add((s, ns.ilxtr.numberOfOccurrences, rdflib.Literal(count))),
              g.add((s, ns.rdf.type, type_)))
             for s, count in counts.items()
             if g.namespace_manager.qname(s).split(':')[0]
             not in ('rdf', 'rdfs', 'owl')]

        rows = [[g.namespace_manager.qname(s),
                 g.namespace_manager.qname(next(g[s:ns.rdf.type:])),
                 o.toPython()]
                for s, o in g[:ns.ilxtr.numberOfOccurrences:]]

        g = OntGraph(namespace_manager=ontres.graph.namespace_manager)  # don't include counts in export
        m = sheets.Mis()
        g.populate_from_triples(m.triples())
        olr = pauth.get_path('ontology-local-repo')
        if olr.exists():
            g.write(path=olr / 'ttl/sparc-mis-helper-gen.ttl')  # FIXME hardcoded and non-obvious

        header = [['curie', 'type', 'count']]
        return header + sorted(rows, key=lambda r:-r[-1])  # FIXME only Report has access to -C :/


class Report:

    @property
    def _sort_key(self):
        if self.options.sort_count_desc:
            return lambda kv: -kv[-1]
        else:
            return lambda kv: kv

    def __init__(self, *args, **kwargs):
        from sparcur.curation import ExporterSummarizer
        self.ExporterSummarizer = ExporterSummarizer
        from sparcur.curation import Integrator, DatasetObject  # FIXME
        self.Integrator = Integrator
        self.DatasetObject = DatasetObject

        super().__init__(*args, **kwargs)

    _all_anno_tags = (
        'protc:input',  # this is slower now
        'protc:input-instance',
        'protc:implied-input',
        'protc:aspect',
        'protc:implied-aspect',
        'protc:executor-verb',
        'protc:parameter*',
        'protc:invariant',
        'protc:output',
        'protc:black-box',
        'protc:black-box-component',
        'protc:objective*',
        'protc:*measure',
        'protc:symbolic-input',
        'protc:symbolic-output',
    )

    def all(self):
        #self.access()  # must be on live data
        # TODO add a prov report for when this was last run etc.
        self.contributors()
        self.filetypes()
        self.samples()
        self.subjects()
        self.samples_values()
        self.subjects_values()
        self.completeness()
        self.keywords()
        self.size()
        self.overview()
        #self.test()  # just a test
        #self.errors()  # TODO too much for now
        #self.pathids()  # too big/slow
        if hasattr(self, 'anchor'):  # FIXME hack
            # branch used to detect if we have a chance of finding the xml export
            # xml ir is not exported with all the rest at the moment
            self.mbf()

        self.terms()
        self.hubmap()
        self.hubmap_anatomy()
        self.mis()
        self.milestones()
        for tag in self._all_anno_tags:
            # FIXME hack around broken anno_tags impl
            self.anno_tags(tag)

    def access(self, ext=None):
        """ Report on datasets that are in the master sheet but which the
            automated pipelines do not have access to. """
        if self.anchor.id == 'N:organization:fake-organization-id':
            # FIXME need a way to generalize/fix this so that organization
            # can vary ...
            print('Not in correct organization for running access report.')
            sys.exit(9999)

        if self.options.server:
            def fmt(d): return hfn.atag(d.uri_human, d.id, new_tab=True)
        else:
            def fmt(d): return d.uri_human

        if self.options.server:
            uri_human = sheets.Master._uri_human()
            def fmtmc(s): return hfn.atag(uri_human, s, new_tab=True)
        else:
            def fmtmc(s): return s

        o = sheets.Overview()
        remote = self.anchor.remote
        remote_datasets = remote.children

        master_sheet_ids = o.dataset_ids()
        bf_api_ids = set(d.id for d in remote_datasets)

        missing_from_api = sorted(master_sheet_ids - bf_api_ids)
        missing_datasets = [self.BlackfynnRemote(i, local_only=True)
                            for i in missing_from_api]
        missing_uris = [fmt(d) for d in missing_datasets]
        rows = [['', ''],
                [fmtmc('Master Count'), len(master_sheet_ids)],
                ['BF API Count', len(bf_api_ids)],
                ['No API Count', len(missing_from_api)],
                *[[i, u] for i, u in zip(missing_from_api, missing_uris)]
        ]
        return self._print_table(rows, title='Access Report', ext=ext)

    @sheets.Reports.makeReportSheet('id')
    def contributors(self, ext=None):
        data = self._data_ir()

        datasets = data['datasets']
        unique = {c['id']:c for d in datasets
                  if 'contributors' in d
                  for c in d['contributors']}
        contribs = sorted(unique.values(),
                          key=lambda c: (c['last_name']  # pYthON dOEsNt hAEV mUlTiLiNE lAmBDaS
                                         if 'last_name' in c else
                                         (c['contributor_name'] if
                                          'contributor_name' in c else
                                          'zzzzzzzzzzzzzzzzzzzzzzzzzzzz')))
        #contribs = sorted((dict(c) for c in
                           #set(frozenset({k:tuple(v) if isinstance(v, list) else
                                          #(frozenset(v.items()) if isinstance(v, dict) else v)
                                          #for k, v in c.items()}.items())
                               #for d in datasets
                               #if 'contributors' in d
                               #for c in d['contributors']
                               #if not log.info(lj(c)))),
                          #key=lambda c: c['last_name'] if 'last_name' in c else c['name'])

        if self.options.debug:
            breakpoint()

        rows = [['id', 'last', 'first', 'PI', 'No Orcid']] + [[
            c['id'],
            c['last_name'] if 'last_name' in c else '',
            c['first_name'] if 'first_name' in c else '',
            'x' if 'contributor_role' in c and 'PrincipalInvestigator' in c['contributor_role'] else '',
            'x' if 'orcid' not in c['id'] else '']
            for c in contribs]

        return self._print_table(rows,
                                 title='Contributors Report',
                                 ext=ext)

    def tofetch(self, dirs=None, ext=None):
        if dirs is None:
            dirs = self.options.directory
            if not dirs:
                dirs.append(self.cwd.as_posix())

        data = []

        def dead(p):
            raise ValueError(p)

        for d in dirs:
            if not Path(d).is_dir():
                continue  # helper files at the top level, and the symlinks that destory python
            path = Path(d).resolve()
            paths = path.rchildren #list(path.rglob('*'))
            path_meta = {p:p.cache.meta if p.cache else dead(p) for p in paths
                         if p.suffix not in ('.swp',)}
            outstanding = 0
            total = 0
            tf = 0
            ff = 0
            td = 0
            uncertain = False  # TODO
            for p, m in path_meta.items():
                #if p.is_file() and not any(p.stem.startswith(pf) for pf in self.spcignore):
                if p.is_file() or p.is_broken_symlink():
                    s = m.size
                    if s is None:
                        uncertain = True
                        continue

                    tf += 1
                    if s:
                        total += s

                    #if '.fake' in p.suffixes:
                    if p.is_broken_symlink():
                        ff += 1
                        if s:
                            outstanding += s

                elif p.is_dir():
                    td += 1

            data.append([path.name,
                         aug.FileSize(total - outstanding),
                         aug.FileSize(outstanding),
                         aug.FileSize(total),
                         uncertain,
                         (tf - ff),
                         ff,
                         tf,
                         td])

        formatted = [[n, l.hr, o.hr, t.hr if not u else '??', lf, ff, tf, td]
                     for n, l, o, t, u, lf, ff, tf, td in
                     sorted(data, key=lambda r: (r[4], -r[3]))]
        rows = [['Folder', 'Local', 'To Retrieve', 'Total', 'L', 'R', 'T', 'TD'],
                *formatted]

        return self._print_table(rows, title='File size counts', ext=ext)

    # TODO generator issue
    #@sheets.Reports.makeReportSheet('suffix', 'mimetype', 'magic_mimetype')
    def filetypes(self, ext=None):
        key = self._sort_key
        paths = self.paths if self.paths else (self.cwd,)
        paths = [c for p in paths for c in p.rchildren if not c.is_dir()]
        rex = re.compile('^\.[0-9][0-9][0-9A-Z]$')
        rex_paths = [p for p in paths if re.match(rex, p.suffix)]
        paths = [p for p in paths if not re.match(rex, p.suffix)]

        def count(thing):
            return sorted([(k if k else '', v) for k, v in
                            Counter([getattr(f, thing)
                                     for f in paths]).items()], key=key)

        each = {t:count(t) for t in ('suffix', 'mimetype', '_magic_mimetype')}
        each['suffix'].append((rex.pattern, len(rex_paths)))

        for title, rows in each.items():
            yield self._print_table(((title, 'count'), *rows),
                                    title=title.replace('_', ' ').strip())

        all_counts = sorted([(*[m if m else '' for m in k], v) for k, v in
                             Counter([(f.suffix, f.mimetype, f._magic_mimetype)
                                      for f in paths]).items()], key=key)

        header = ['suffix', 'mimetype', 'magic mimetype', 'count']
        return self._print_table((header, *all_counts),
                                 title='All types aligned (has duplicates)',
                                 ext=ext)

    def _s(self, dict_key, ext=None):
        data = self._data_ir()
        datasets = data['datasets']
        key = self._sort_key
        # FIXME we need the blob wrapper in addition to the blob generator
        # FIXME these are the normalized ones ...
        s_headers = tuple(
            uk for dataset_blob in datasets
            if dict_key in dataset_blob  # FIXME inputs?
            # unique so that we don't bias by the number of specimens
            # in a given study, there is an issue with sparse
            # reporting over different specimens in the same study,
            # but not much we can do about that
            for uk in set(
                    k for s_blob in dataset_blob[dict_key]
                    for k in s_blob))
        counts = tuple(kv for kv in sorted(Counter(s_headers).items(),
                                            key=key))

        index_col_name = 'Column Name'
        if ext is None:
            index_col_name += f' unique = {len(counts)}'
        rows = ((index_col_name, '#'), *counts)
        return self._print_table(rows,
                                 title=f'{dict_key.capitalize()} Report',
                                 ext=ext)


    @sheets.Reports.makeReportSheet('column_name', sheet_name='samples')
    def samples(self, ext=None):
        return self._s('samples', ext=ext)

    @sheets.Reports.makeReportSheet('column_name', sheet_name='subjects')
    def subjects(self, ext=None):
        return self._s('subjects', ext=ext)

    def _s_values(self, dict_key, ext=None, skip_keys=('subject_id', 'sample_id', 'primary_key')):
        data = self._data_ir()
        datasets = data['datasets']

        def detype(_v):
            if isinstance(_v, dict) and '@value' in _v:
                # jsonld
                return (_v['@value'],)

            return ((_v.asJson(),) if hasattr(_v, 'asJson') else
                    ((_v.identifier,) if hasattr(_v, 'identifier')
                     else (_v if is_list_or_tuple(_v) else (_v,))))

        def anydict(_v):
            if isinstance(_v, dict):
                return [_v]
            elif not is_list_or_tuple(_v):
                return False

            oops = []
            JApplyRecursive(get_nested_by_type, _v, dict, collect=oops)
            return oops

        dd = defaultdict(set)
        _ = [
            dd[k].add(v) for dataset_blob in datasets
            if dict_key in dataset_blob
            for samples_blob in dataset_blob[dict_key]
            for k, _v in samples_blob.items()
            if k not in skip_keys and not anydict(_v)
            for v in detype(_v)
        ]
        _ = [
            dd[k].add(v) for dataset_blob in datasets
            if dict_key in dataset_blob
            for samples_blob in dataset_blob[dict_key]
            for _k, __v in samples_blob.items()
            if _k not in skip_keys and anydict(__v)
            for blob in anydict(__v)
            for k, _v in blob.items()
            if k not in skip_keys
            for v in detype(_v)
        ]

        vd = dict(dd)
        header = sorted(vd, key=lambda k:len(vd[k]))
        cols = [[k, *sorted([str(v) for v in vd[k]])] for k in header]
        rows = list(zip_longest(*cols, fillvalue=None))
        # FIXME something about the way we process updates is broken
        # in cases where we try to replace the whole sheet, the end
        # result is that if you delete all the content you get an
        # index out of range error, and in other cases you wind up
        # having multiple reports on top of eachother in the wrong
        # columns
        return self._print_table(rows,
                                 title=f'{dict_key.capitalize()} Values Report',
                                 ext=ext)

    @sheets.Reports.makeReportSheet(sheet_name='samples-values')
    def samples_values(self, ext=None):
        return self._s_values('samples', ext=ext)

    @sheets.Reports.makeReportSheet(sheet_name='subjects-values')
    def subjects_values(self, ext=None):
        return self._s_values('subjects', ext=ext)

    @sheets.Reports.makeReportSheet('id')  # FIXME vs path
    def completeness(self, ext=None):
        if self.options.raw:
            sb = self._data_ir()
            raw = (self.summary._completeness(d) for d in sb['datasets'])
        else:
            from sparcur import export as ex
            datasets = self._data_ir()['datasets']
            raw = [self.ExporterSummarizer._completeness(data) for data in datasets]

        def rformat(i, si, ci, ei, name, id, award, organ):
            if self.options.server and isinstance(ext, types.FunctionType):
                # rsurl = 'https://projectreporter.nih.gov/reporter_searchresults.cfm'
                dataset_dash_url = self.url_for('route_datasets_id', id=id)
                errors_url = self.url_for('route_reports_errors_id', id=id)
                si = hfn.atag(errors_url + '#submission', si)
                ci = hfn.atag(errors_url + '#curation', ci)
                ei = hfn.atag(errors_url + '#total', ei)
                name = hfn.atag(dataset_dash_url, name)
                id = hfn.atag(dataset_dash_url, id[:10] + '...')
                award = (
                    hfn.atag(('https://scicrunch.org/scicrunch/data/source/'
                              f'nif-0000-10319-1/search?q={award}'), award)
                    if award else 'MISSING')
                organ = organ if organ else ''
                if isinstance(organ, list) or isinstance(organ, tuple):
                    organ = ' '.join([o.atag() for o in organ])
                if isinstance(organ, OntTerm):
                    organ = organ.atag()
            else:
                award = award if award else ''
                organ = ((repr(organ) if isinstance(organ, OntTerm) else organ)
                         if organ else '')
                if isinstance(organ, list):
                    organ = ' '.join([repr(o) for o in organ])

            return (i + 1, si, ci, ei, name, id, award, organ)

        rows = [('', 'SI', 'CI', 'EI', 'name', 'id', 'award', 'organ')]
        rows += [rformat(i, *cols) for i, cols in enumerate(sorted(
            raw, key=lambda t: (t[0], t[1], t[5] if t[5] else 'z' * 999, t[3])))]

        return self._print_table(rows,
                                 title='Completeness Report',
                                 ext=ext)

    @sheets.Reports.makeReportSheet()
    def keywords(self, ext=None):
        data = self._data_ir()
        datasets = data['datasets']
        _rows = [sorted(set(dataset_blob.get('meta', {}).get('keywords', [])),
                        key=lambda v: -len(v))
                    for dataset_blob in datasets]
        rows = [list(r) for r in sorted(
            set(tuple(r) for r in _rows if r),
            key = lambda r: (len(r), tuple(len(c) for c in r if c), r))]
        header = [[f'{i + 1}' for i, _ in enumerate(rows[-1])]] if rows else []
        rows = header + rows
        return self._print_table(rows,
                                 title='Keywords Report',
                                 ext=ext)


    @sheets.Reports.makeReportSheet('id')
    def size(self, ext=None):
        data = self._data_ir()
        if 'export_project_path' in data['prov']:
            project_path = PurePath(data['prov']['export_project_path'])
        else:
            # best guess
            project_path = PurePath(
                '/var/lib/sparc/files/sparc-datasets/',
                PennsieveId(data['id']).uuid,
                'dataset')

        rows = [['path', 'id', 'sparse', 'dirs', 'files', 'size', 'hr'],
                *sorted([[(project_path / m['folder_name']).name
                          if 'folder_name' in m else '',
                          d['id'],
                          'x' if 'sparse' in m and m['sparse'] else '?'] +
                         [int(m[k]) if k in m else -1  # float('-inf')
                          for k in ['dirs', 'files', 'size']] +
                         [aug.FileSize(m['size']).hr
                          if 'size' in m else -1  # float('-inf')
                          ] for d in data['datasets'] for m in [d['meta']]],
                        key=lambda r: -r[-2])]

        return self._print_table(rows, title='Size Report',
                                 align=['l', 'l', 'r', 'r', 'r', 'r', 'r'],
                                 ext=ext)

    @sheets.Reports.makeReportSheet('id')
    def overview(self, ext=None):
        data = self._data_ir()
        datasets = data['datasets']
        hrm = {'award': ['meta', 'award_number'],
               'id': ['id'],
               'name': ['meta', 'folder_name'],
               'errors': ['status', 'error_index'],
               'updated': ['status', 'updated'],
               'published': ['meta', 'doi'],
               'milestone_completion_date': ['submission',
                                             'milestone_completion_date'],
               # TODO
               # subject_count
               # sample_count
               # contributor_count
               }
        header = tuple(hrm)  # TODO counts ?
        paths = tuple(hrm.values())
        def getl(d):
            return [adops.get(d, path, on_failure='') for path in paths]
        _rows = sorted([getl(d) for d in datasets],
                       key=lambda r: (r[-2].asStr()
                                      if isinstance(r[-2], idlib.Stream) else
                                      r[-2],
                                      print(r),
                                      r[0],
                                      r[-3],
                                      r[-4],
                                      ))
        rows = [header] + _rows
        return self._print_table(rows, title='Dataset Report',
                                 align=['l', 'l', 'l', 'r', 'r', 'r', 'r'],
                                 ext=ext)

    def test(self, ext=None):
        rows = [['hello', 'world'], [1, 2]]
        return self._print_table(rows, title='Report Test', ext=ext)


    #@sheets.Reports.makeReportSheet('id')  # TODO bad return format right now
    def errors(self, *, id=None, ext=None):
        data = self._data_ir()
        datasets = data['datasets']

        if self.cwd != self.anchor:
            id = self.cwd.cache.dataset.id

        if id is not None:
            if not id.startswith('N:dataset:'):
                return []

            def pt(rendered_table, title=None):
                """ passthrough """
                return rendered_table

            import htmlfn as hfn
            for dataset_blob in datasets:
                if dataset_blob['id'] == id:
                    dso = self.DatasetObject.from_json(dataset_blob)
                    title = f'Errors for {id}'
                    urih = dataset_blob['meta']['uri_human']
                    formatted_title = (
                        hfn.h2tag(f'Errors for {hfn.atag(urih, id)}<br>\n') +
                        (hfn.h3tag(dataset_blob['meta']['title']
                                   if 'title' in dataset_blob['meta'] else
                                   dataset_blob['meta']['folder_name'])))
                    log.info(list(dataset_blob.keys()))
                    errors = list(dso.errors)
                    return [(self._print_table(e.as_table(), ext=pt))
                            for e in errors], formatted_title, title
        else:
            pprint.pprint(
                sorted([(d['meta']['folder_name'],
                         [e['message'] for path, e in get_all_errors(d)])
                        for d in datasets],
                       key=lambda ab: (bool(ab), -len(ab[-1]) if ab else 0)))

    def pathids(self, ext=None):
        base = self.project_path.parent
        rows = [['path', 'id']] + sorted([c.relative_to(base), c.cache.id]#, c.cache.uri_api, c.cache.uri_human]
                                         # slower to include the uris
                                         for c in chain((self.cwd,),
                                                        self.cwd.rchildren)
        )
        return self._print_table(rows, title='Path identifiers', ext=ext)

    #@sheets.Reports.makeReportSheet('id')  # TODO bad return format right now
    def mbf(self, ext=None):
        et = tuple()
        from sparcur.extract import xml as exml

        def settype(mimetype):
            return {exml.ExtractMBF.mimetype: 'MBF Metadata',
                    exml.ExtractNeurolucida.mimetype: 'Neurolucida',}[mimetype]

        if self.options.raw:
            blob_ir = self.parent.export()
        else:
            from sparcur import export as ex
            blob_ir = self._export(ex.ExportXml).latest_ir  # FIXME need to load?

        mbf_types = tuple(c.mimetype for c in
                          (exml.ExtractMBF, exml.ExtractNeurolucida))
        if self.options.unique:
            key = lambda p: (p[2], not p[0], p[1].lower(), p[1])
        else:
            key = lambda p: (p[2], p[3], not p[0], p[1].lower(), p[1])

        all_conts = sorted(set(((OntId(c['id_ontology'])
                                 if 'id_ontology' in c else
                                 ''),
                                c['name'],
                                settype(metadata_blob['type']),
                                *((dataset_xml['dataset_id'],)
                                  if not self.options.unique else et))
                               for dataset_xml in blob_ir.values()
                               if dataset_xml['type'] == 'all-xml-files'  # FIXME
                               for metadata_blob in dataset_xml['xml']
                               if metadata_blob['type'] in mbf_types and
                               'contours' in metadata_blob['contents']
                               for c in metadata_blob['contents']['contours']),
                           key=key)
        header = [['id', 'name', 'metadata source', 'dataset']]
        if self.options.unique:
            header[0] = header[0][:-1]

        return self._print_table(header + all_conts,
                                 title='Unique MBF contours',
                                 ext=ext)

    @property
    def _graph(self):
        # FIXME cache these results and only recompute if latest changes?
        if self.options.raw:
            graph = self.summary.triples_exporter.graph
        else:
            #from sparcur import export as ex  # FIXME very slow to import
            #graph = OntGraph()
            #self._export(ex.Export).latest_export_ttl_populate(graph)
            graph = self._ttlfile.graph

        return graph

    def terms(self, ext=None):
        if self.options.hubmap:
            return self.hubmap(ext=ext)
        elif self.options.hubmap_anatomy:
            return self.hubmap_anatomy(ext=ext)

        # anatomy
        # cells
        # subcelluar
        import rdflib
        graph = self._graph
        objects = set()
        skipped_prefixes = set()
        for t in graph:
            for e in t:
                if (isinstance(e, rdflib.URIRef) and
                    not e.startswith('info:') and
                    not e.startswith('doi:')):
                    oid = OntId(e)
                    if oid.prefix in want_prefixes:
                        objects.add(oid)
                    elif oid.prefix is not None:
                        skipped_prefixes.add(oid.prefix)

        if self.options.server and isinstance(ext, types.FunctionType):
            def reformat(ot):
                return [ot.label
                        if hasattr(ot, 'label') and ot.label else
                        '',
                        ot.atag(curie=True)]

        else:
            def reformat(ot):
                return [ot.label
                        if hasattr(ot, 'label') and ot.label else
                        '',
                        ot.curie]

        log.info(' '.join(sorted(skipped_prefixes)))
        objs = [OntTerm(o) if o.prefix not in ('TEMP', 'sparc') or
                o.prefix == 'TEMP' and o.suffix.isdigit() else
                o for o in objects]
        term_sets = {title:[o for o in objs if o.prefix == prefix]
                     for prefix, title in
                     (('NCBITaxon', 'Species'),
                      ('UBERON', 'Anatomy and age category'),  # FIXME
                      ('FMA', 'Anatomy (FMA)'),
                      ('PATO', 'Qualities'),
                      ('tech', 'Techniques'),
                      ('unit', 'Units'),
                      ('sparc', 'MIS terms'),
                      ('TEMP', 'Suggested terms'),
                     )}

        term_sets['Other'] = set(objs) - set(ot for v in term_sets.values()
                                             for ot in v)

        for title, terms in term_sets.items():
            header = [['Label', 'CURIE']]
            rows = header + [reformat(ot) for ot in
                            sorted(terms,
                                   key=lambda ot: (ot.prefix, ot.label.lower()
                                                   if hasattr(ot, 'label') and ot.label else ''))]

            yield self._print_table(rows, title=title, ext=ext)

    @idlib.utils.cache_result
    def _hubmap(self):
        graph = self._graph
        ir = self._data_ir()
        rows = self._hubmap_terms(graph, ir)
        return rows

    @sheets.Reports.makeReportSheet()
    def hubmap(self, ext=None):
        title = 'SPARC Terms for HuBMAP'
        rows = self._hubmap()
        rows = expand_label_curie(rows)
        rows = [['Species', 'Species Id', 'Term', 'Term Id']] + rows
        return self._print_table(rows, title=title, ext=ext)

    @sheets.Reports.makeReportSheet()
    def hubmap_anatomy(self, ext=None):
        title = 'SPARC Anatomical Terms for HuBMAP'
        rows = self._hubmap()
        anat = [r for r in rows if r[1].prefix in ('UBERON', 'FMA', 'ILX')]
        rows = expand_label_curie(anat)
        rows = [['Species', 'Species Id', 'Anatomy', 'Anatomy Id']] + rows
        return self._print_table(rows, title=title, ext=ext)

    @staticmethod
    def _hubmap_terms(graph, ir):
        import rdflib
        def collect_id_ontology(obj, key, *args, path=None, collect=tuple()):
            lc = len(collect)
            get_nested_by_key(obj, key, *args, path=path, collect=collect)
            if lc + 1 == len(collect):
                collect[-1] = (collect[-1], tuple(path[:2]))

            return obj


        def get_all_types(obj, *args, path=None, collect=tuple()):
            collect.add(type(obj))
            return obj


        def collect_types(obj, type, *args, path=None, collect=tuple()):
            lc = len(collect)
            get_nested_by_type(obj, type, *args, path=path, collect=collect)
            if lc + 1 == len(collect):
                collect[-1] = (collect[-1], tuple(path[:2]))

            return obj

        # ttl
        # may have to run with compute_qname once to avoid an error?
        sigh = [(graph.namespace_manager.compute_qname(s), o.toPython())
        for s, o in graph[:isAbout:] if isinstance(o, rdflib.URIRef)]

        asdf = sorted(set([(graph.namespace_manager.qname(s), o.toPython())
                for s, o in graph[:isAbout:] if isinstance(o, rdflib.URIRef)] + [
                        (graph.namespace_manager.qname(s), o.toPython())
        for s, o in graph[:TEMP.involvesAnatomicalRegion:] if isinstance(o, rdflib.URIRef)]))

        qq = defaultdict(list)
        for s, o in asdf:
            qq[s].append(OntTerm(o))

        # json
        collect = []
        _ = JApplyRecursive(get_nested_by_type, ir, str, collect=collect)
        eff = sorted(set(collect), key=lambda k:len(k))

        collect = []
        _ = JApplyRecursive(get_nested_by_type, ir, OntTerm, collect=collect)
        terms = sorted(set(collect))

        collect = set()
        _ = JApplyRecursive(get_all_types, ir, collect=collect)
        types = collect

        collect = []
        _ = JApplyRecursive(collect_id_ontology, ir, 'id_ontology', collect=collect)
        _from_mbf = sorted(set(collect))
        from_mbf = [(OntTerm(t), p) for t, p in _from_mbf]

        collect = []
        _ = JApplyRecursive(collect_types, ir, OntTerm, collect=collect)
        term_dsps = sorted(set(collect))

        dids = {('datasets', i): d['id'][2:] for i, d in enumerate(ir['datasets'])}

        for t, p in from_mbf + term_dsps:
            id = dids[p]
            qq[id].append(t)

        banned = set(OntTerm(c, label=l)
                    for c, l in (('UBERON:0000025', 'tube'),
                                 ('UBERON:0000479', 'tissue') ,
                                 ('UBERON:0003978', 'valve'),
                                 ('UBERON:0000403', 'scalp'),  # guessing scalple
                                 ('UBERON:0001021', 'nerve'),
                                 ('UBERON:0018707', 'bladder organ'),
                                 ('UBERON:4200215', 'suture'),
                                 ('UBERON:0003129', 'skull'),
                                 ('UBERON:0001753', 'cementum'),  # guessing cement
                                 ('UBERON:0003148', 'obsolete sclerite'),  # plates apparently >_<
                                ))
        zz = defaultdict(set)
        for terms in qq.values():
            spec = None
            anat = []
            for term in terms:
                if term in banned:
                    continue

                if term.prefix == 'NCBITaxon':
                    spec = term
                else:
                    anat.append(term)

            zz[spec].update(anat)

        # structure
        # species organ
        # linked via dataset

        rows = sorted([[s, t] for s, ts in zz.items() for t in ts])
        return rows

    @idlib.utils.cache_result
    def _annos(self):
        cache_file = self.options.hypothesis_cache_file
        if cache_file is None:
            group_id = auth.get('hypothesis-group')
            if group_id is None:
                # FIXME better error message and add an
                # orthognoal check on stuff like this
                raise ValueError('no hypothesis group')
            cache_file = Path(hyp.group_to_memfile(group_id + 'sparcur'))
            get_annos = hyp.Memoizer(memoization_file=cache_file, group=group_id)
            get_annos.api_token = auth.get('hypothesis-api-key')
            annos = get_annos()
        else:
            group_name = self.options.hypothesis_group_name  # there is a default value
            group_id = (auth.user_config.secrets('hypothesis', 'group', group_name)
                if group_name else auth.get('hypothesis-group'))
            get_annos = hyp.AnnoReader(memoization_file=cache_file, group=group_id)
            annos, last_sync_updated = get_annos.get_annos_from_file()

        self._annos_cache_file = cache_file
        return annos

    @idlib.utils.cache_result
    def _annos_bpc(self):
        from protcur import document as ptcdoc
        annos = self._annos()
        bannos = [ptcdoc.Annotation(a) for a in annos]  # better annos
        pool = ptcdoc.Pool(bannos)
        anno_counts = ptcdoc.AnnoCounts(pool)
        return annos, bannos, pool, anno_counts

    @idlib.utils.cache_result
    def _helpers(self):
        annos = self._annos()
        return [hyp.HypothesisHelper(a, annos) for a in annos]

    @idlib.utils.cache_result
    def _protc(self):
        from pysercomb.parsers import racket
        from protcur.analysis import protc
        annos = self._annos()
        return [protc(a, annos) for a in annos]

    def _protc_input(self, hrmd, tags, protc, ext):
        from pysercomb import exceptions as pexc
        from pysercomb.parsers import racket
        import pysercomb.pyr.units as pyru
        pyru.Hyp.bindImpl(None, HypothesisAnno=protc.byId)
        protcur_interpreter = pyru.Protc()
        def more(k, v):  # FIXME these can actually be parsed in parallel
            count = len(v)
            link = v[0].shareLink if count == 1 else ''
            facet = 'TODO'
            autos = set()
            for protc_helper in v:
                ok, expr, rest = racket.exp(repr(protc_helper))
                #log.critical(v)  # FIXME this induces an infinite loop I think due to a missing child
                #log.critical(expr)
                if expr is None or isinstance(expr, str):
                    continue
                try:
                    hrm = protcur_interpreter(expr)
                except pexc.ParseFailure as e:
                    log.error(e)
                    continue

                if isinstance(hrm._value, pyru.Term):
                    autos.add(hrm._value)

                continue

            auto_mapping = ';'.join([f'{bb.curie}|{bb.label}' for bb in autos])
            return count, link, facet, auto_mapping

        header = [['tag', 'value', 'text', 'exact', 'count', 'link', 'facet', 'auto-mapping']]
        #rows = [[(*k, ' '.join([a.shareLink for a in v]))] for k, v in sorted(hrm.items())]
        rows = [[*k, *more(k, v)]
                for k, v in sorted(list(hrmd.items()))]

        if self.options.sort_count_desc:
            rows = sorted(rows, key=lambda r: -r[-4])

        return self._print_table(header + rows,
                                 title=f'Annos for {" ".join(tags)}',
                                 ext=ext)

    def _process_protc(self, tags, matches, ext):
        from protcur.analysis import protc, ParameterValue
        from pysercomb.pyr import units as pyru
        from pysercomb.parsers import racket
        pyru.Term._OntTerm = OntTerm  # the tangled web grows ever deeper :x

        lowernorm = ('protc:aspect',
                     'protc:implied-aspect',
                     'protc:executor-verb',
                     'protc:parameter*',
                     'protc:invariant',
                     )

        self._protc()
        # FIXME cannot breakpoint with this protc instances around ...
        pm = [protc.byId(a.id) for _, a in matches]

        # testing
        if 'protc:input' in tags:
            input = pm[13]
            invar = next(next(input.children).children)
            paramparser = pyru.ParamParser()
            param = invar.parameter()
            if isinstance(param, ParameterValue):  # FIXME how the heck ... it was a hack in analysis iirc
                tv = param.v
            else:
                tv = racket.sexp(param)[1]
            tp = invar._parameter[1]
            assert tv == tp, breakpoint()
            hrm = paramparser(tv)

        hrm = defaultdict(list)
        def denone(t): return tuple('' if e is None else e for e in t)
        for p in pm:
            pt = [t for t in p.tags if t in tags]
            if pt:
                hrm[denone((
                    pt[0],
                    normalize_exact(
                        p.value,
                        lower=anyMembers(tags, *lowernorm)),
                    normtext(
                        p.text,
                        lower=anyMembers(tags, *lowernorm)),
                    normalize_exact(
                        p.exact,
                        lower=anyMembers(tags,
                                            *lowernorm))))].append(p)
            else:
                log.warning(f'WAT {p.tags}')

        hrmd = dict(hrm)
        if 'protc:executor-verb' in tags:  # woo past tense
            ends_with_e = {h[1]:h for h, p in hrmd.items() if h[1].endswith('e')}
            others = {h[1]:h for h, p in hrmd.items()
                    if not h[1].endswith('ed') and not h[1].endswith('e')}
            for k, p in list(hrmd.items()):
                _, v, _, _ = k
                if v.endswith('ed'):
                    ewe = v[:-1]
                    if ewe in ends_with_e:
                        nk = ends_with_e[ewe]
                    elif v[:-2] in others:
                        nk = others[v[:-2]]
                    else:
                        continue

                    hrmd[nk].extend(hrmd.pop(k))

            # this has to come after due to remapping that happens above first
            wev = sheets.WorkingExecVerb()
            value_to_key, create = wev.condense()
            for _key in create:
                if _key not in hrmd:
                    hrmd[_key] = []

            for k, p in list(hrmd.items()):
                _, v, _, _ = k
                if v in value_to_key:
                    nk = value_to_key[v]
                    hrmd[nk].extend(hrmd.pop(k))

        if 'protc:input' in tags or 'protc:black-box' in tags or 'protc:black-box-component' in tags:
            return self._protc_input(hrmd, tags, protc, ext)

        def more(k, v):
            count = len(v)
            link = v[0].shareLink if count == 1 else ''
            facet = 'TODO'
            return count, link, facet

        header = [['tag', 'value', 'text', 'exact', 'count', 'link', 'facet']]
        #rows = [[(*k, ' '.join([a.shareLink for a in v]))] for k, v in sorted(hrm.items())]
        rows = [[*k, *more(k, v)]
                for k, v in sorted(hrmd.items())]

        if self.options.sort_count_desc:
            rows = sorted(rows, key=lambda r: -r[-3])

        return self._print_table(header + rows,
                                    title=f'Annos for {" ".join(tags)}',
                                    ext=ext)

    def _by_tags(self, *tags, links=True, ext=None):
        annos, bannos, pool, anno_counts = self._annos_bpc()

        self._helpers()
        process_protc = anyMembers(tags,
                                   'protc:executor-verb',
                                   'protc:input',
                                   'protc:aspect',
                                   'protc:input-instance',
                                   'protc:implied-aspect',
                                   'protc:implied-input',
                                   'protc:parameter*',
                                   'protc:invariant',
                                   'protc:output',
                                   'protc:black-box',
                                   'protc:black-box-component',
                                   )

        # note that HypothesisHelper.byTags is an AND search not and OR search
        matches = [(normalize_exact(a.exact, lower=True), a)
                   for tag in tags for a in hyp.HypothesisHelper.byTags(tag)]

        if process_protc:
            return self._process_protc(tags, matches, ext)

        md = defaultdict(list)
        for e, a in matches:
            md[e].append(a)

        def gtag(e):  # FIXME memoize/cache?
            return ' '.join(sorted(set(t for a in md[e] for t in a.tags
                                       if t in tags)))

        def link(e): pass
        rows = sorted([[gtag(e),  # tag
                        e,        # key
                        c,        # count
                        (' '.join([a.htmlLink  # links
                                   if self.options.uri_html else
                                   a.shareLink for a in md[e]]
                                  if links else '')),
                        '',       # docs
                        '',       # exacts
                        ]
                       for e, c in Counter([normalize_exact(e)
                                            for e, a in matches]).most_common()],
                      key=lambda ab: (ab[0], -ab[2], ab[1].lower()))

        header = [['tag', 'key', 'count', 'links', 'docs', 'exacts']]
        return self._print_table(header + rows,
                                 title=f'Annos for {" ".join(tags)}',
                                 ext=ext)

    @sheets.AnnoTags.makeReportSheet()
    def anno_tags(self, *tags, links=False, ext=None):
        if not tags:
            tags = self.options.tag

        links = links or self.options.uri or self.options.uri_api

        if list(tags) == ['all']:
            out = []
            for tag in self._all_anno_tags:
                o = self._by_tags(tag, links=links, ext=ext)
                out.append(o)

            return out
        else:
            return self._by_tags(*tags, links=links, ext=ext)

    @property
    @idlib.utils.cache_result
    def _ttlfile(self):
        # TODO uri/path reconciliation 
        if self.options.ttl_file:
            ontres = OntRes.fromStr(self.options.ttl_file)
        else:
            # FIXME make this go away and derive it from git
            iri = 'https://cassava.ucsd.edu/sparc/preview/exports/curation-export.ttl'
            ontres = OntResIri(iri)

        if self.options.ttl_compare:
            ontres_compare_to = OntRes.fromStr(self.options.ttl_compare)
        else:
            ontres_compare_to = None

        return TtlFile(ontres, ontres_compare_to)

    def changes(self):
        a, r, c = self._ttlfile.changes()
        for n, t in zip(('Added', 'Removed', 'Changed'), (a, r, c)):
            print(n + ' ' + ('-' * 40))
            pprint.pprint(t)
            print(('-' * (len(n) + 1)) + ('-' * 40))

    @sheets.Reports.makeReportSheet('curie')
    def mis(self, ext=None):
        tout = self._ttlfile.mis(ext=ext)
        return self._print_table(tout,
                                 title=f'MIS predicates',
                                 ext=ext)

    @sheets.Reports.makeReportSheet('dataset')
    def milestones(self, ext=None):
        tout = self._ttlfile.milestones(ext=ext)
        return self._print_table(tout,
                                 title=f'Dataset milestone completion dates',
                                 ext=ext)

    def _get_protocol_ids(self, blob_data, blob_protcur):
        def tp(i):
            """ not pio at all """
            try:
                if i is not None:
                    return idlib.Pio(i)
            except idlib.exc.MalformedIdentifierError:
                return i

        def ti(i):
            """ can't figure out the integer id """
            try:
                return i.uri_api_int
            except (AttributeError,
                    idlib.exc.MalformedIdentifierError,
                    idlib.exc.RemoteError) as e:
                return i

        def do_deref(d):
            r = d.dereference()
            if r != d.asUri():
                return r
            #try:
            #chain = d._resolution_chain_responses(d.identifier_actionable)
            #except (idlib.exc.ResolutionError, idlib.exc.RemoteError) as e:
            #log.exception(e)

        # dataset blob
        collect = []
        _ = JApplyRecursive(get_nested_by_key,
                            blob_data,
                            'protocol_url_or_doi',
                            asExport=False,
                            collect=collect,
                            skip_keys=('inputs',),)
        _ = None
        blob_dataset = None  # no longer needed and repr issues
        collect = sorted(set(collect), key=lambda i: str(i))
        dois = [c for c in collect if isinstance(c, idlib.Doi)]
        deref = Async()(deferred(do_deref)(d) for d in dois)
        dios = [ti(tp(u)) if u is not None else d for d, u in zip(dois, deref)]

        pios = [c for c in collect if isinstance(c, idlib.Pio)]
        pints = [ti(p) for p in pios]
        pstrs = [p.asStr() for p in pints if p is not None]
        from_datasets = set(pints) | set(dios)

        # protcur blob
        protocols = [b for b in blob_protcur['@graph']
                     if '@type' in b and 'sparc:Protocol' in b['@type']]
        urls_protcur = [b['id'] for b in protocols]

        uios = [tp(u) for u in urls_protcur]
        from_hypothesis = set(uios)

        both = from_hypothesis & from_datasets
        only_hyp = from_hypothesis - from_datasets
        only_dat = from_datasets - from_hypothesis
        either = from_hypothesis | from_datasets
        hrm = [(len(s), [d.identifier if hasattr(d, 'identifier') else d for d in s])
               for s in (both, only_hyp, only_dat)]
        return (either, uios, dois, dios, deref,
                from_datasets, from_hypothesis, protocols)

    @sheets.Reports.makeReportSheet('uri')
    def protocols(self, ext=None):
        def th(i):
            try:
                return i.uri_human
            except (AttributeError,
                    idlib.exc.MalformedIdentifierError,
                    idlib.exc.RemoteError) as e:
                pass

        def tht(i):
            try:
                return i.uri_human_html
            except (AttributeError,
                    idlib.exc.MalformedIdentifierError,
                    idlib.exc.RemoteError) as e:
                pass

        blob_protcur = self._protcur()
        blob_data = self._data_ir(org_id=self.options.project_id)

        (either, uios, dois, dios, deref, from_datasets,
         from_hypothesis, protocols) = self._get_protocol_ids(
             blob_data, blob_protcur)

        header = [['uri',
                   'uri_html',
                   'uri_human',
                   'doi',
                   #'datasets',
                   'exists',
                   'access',
                   'count annos',
                   'count protcur',
                   'is protocols.io',
                   'from datasets',
                   'from hypothes.is',
                   'from protocols.io']]
        li = list(either)
        rows = []
        # FIXME probably better to do this by checking if we have a record of this in protocols first
        for i in li:
            if i is None:
                raise TypeError('sigh')
            access = False
            exists = False
            is_pio = False
            if isinstance(i, idlib.Stream):
                is_pio = isinstance(i, idlib.Pio)
                uri = i
                uri_html = tht(i)
                uri_human = th(i)
                try:
                    if isinstance(i, idlib.Doi):
                        doi = i
                        i.data()
                    else:
                        doi = i.doi
                        exists = True
                        access = True
                except idlib.exc.NotAuthorizedError:
                    exists = True
                    doi = None
                except idlib.exc.RemoteError:
                    doi = None
            else:
                uri = idlib.Uri(i)
                uri_html = None
                uri_human = None
                if i in dios:
                    exists = bool(deref[dios.index(i)])
                    doi = dois[dios.index(i)]
                    access = exists
                    #if not exists:  # need the uris as a backstop for ident
                        #uri = None
                else:
                    doi = None
                    try:
                        uri._resolution_chain_responses(uri.identifier_actionable)
                        access = exists = True
                    except idlib.exc.ResolutionError:
                        pass

            if i in from_hypothesis:
                protocol = protocols[uios.index(i)]
                n_annos = protocol['anno_count']
                n_protcur = protocol['protcur_anno_count']
                if uri_human is None and 'uri_human' in protocol:
                    uri_human = protocol['uri_human']
            else:
                n_annos = 0
                n_protcur = 0

            row = [
                uri,
                uri_html,
                uri_human,
                doi,
                #datasets,
                exists,
                access,
                n_annos,
                n_protcur,
                is_pio,
                i in from_datasets,
                i in from_hypothesis,
                False,
            ]

            rows.append(row)

        def sk(r):
            return (
                r[4],
                r[5],
                r[3] is not None, r[3],
                r[2] is not None, r[2],
                r[0] is not None, r[0],
            )

        srows = sorted(rows, key=sk)

        return self._print_table(header + srows,
                                 title=f'Protocol report',
                                 ext=ext)



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::fetch_annotations.py][fetch_annotations.py]]
from pathlib import Path
from hyputils import hypothesis as hyp
from sparcur.config import auth


def from_group_name_fetch_annotations(group_name):
    """ pull hypothesis annotations from remote to local """
    group_id = auth.user_config.secrets('hypothesis', 'group', group_name)
    cache_file = Path(hyp.group_to_memfile(group_id + 'sparcur'))
    get_annos = hyp.Memoizer(cache_file, group=group_id)
    get_annos.api_token = auth.get('hypothesis-api-key')  # FIXME ?
    annos = get_annos()
    return cache_file  # needed for next phase, annos are not


def from_group_name_cached_annos(group_name):
    group_id = auth.user_config.secrets('hypothesis', 'group', group_name)
    cache_file = Path(hyp.group_to_memfile(group_id + 'sparcur'))
    get_annos = hyp.AnnoReader(cache_file, group=group_id)
    annos = get_annos()
    return annos


def from_user_name_group_name_h(user_name, group_name):
    group_id = auth.user_config.secrets('hypothesis', 'group', group_name)
    h = hyp.HypothesisUtils(username=user_name, group=group_id)
    h.token = auth.user_config.secrets('hypothesis', 'api', user_name)
    return h


def main(hypothesis_group_name=None, **kwargs):
    if hypothesis_group_name is None:
        hypothesis_group_name = 'sparc-curation'

    from_group_name_fetch_annotations(hypothesis_group_name)


if __name__ == '__main__':
    from sparcur.simple.utils import pipe_main
    pipe_main(main)
# fetch_annotations.py ends here



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::validate.py][validate.py]]
import augpathlib as aug
from sparcur import pipelines as pipes
from sparcur.paths import PathL, CacheL
from sparcur.utils import GetTimeNow


def makeValidator(dataset_path, time_now=None):

    if time_now is None:
        time_now = GetTimeNow()

    class CacheX(CacheL):
        def __new__(cls, *args, **kwargs):
            return super().__new__(cls, *args, **kwargs)

    CacheX._bind_flavours()

    class PathX(PathL):
        """ Workaround absense of cache. """

        _cache_class = CacheX

        def __new__(cls, *args, **kwargs):
            return super().__new__(cls, *args, **kwargs)

        # TODO likely will also need to rebind the cache class as well

        #@property
        #def dataset_relative_path(self, __drp=dataset_path):
            #return self.relative_path_from(self.__class__(__drp))

    CacheX._local_class = PathX
    PathX._bind_flavours()

    # XXX monkey patch TODO sigh FIXME DatasetStructure calls Path directly inside
    #PathL.dataset_relative_path = Path.dataset_relative_path

    # must caste before anything else is done so that anchor and
    # datasets are known
    dataset_path = PathX(dataset_path)

    CacheX._dataset_dirs = [CacheX(dataset_path)]
    # FIXME this is very much not ideal because we don't actually want
    # the parent in this case
    CacheX._asserted_anchor = CacheX(dataset_path.parent)

    class context:
        path = dataset_path.resolve()
        id = path.id
        uri_api = path.as_uri()
        uri_human = path.as_uri()

    class lifters:
        id = context.id
        remote = 'local'
        folder_name = context.path.name
        uri_api = context.uri_api
        uri_human = context.uri_human
        timestamp_export_start = time_now.START_TIMESTAMP

        affiliations = lambda *args, **kwargs: None
        techniques = tuple()
        modality = None
        organ_term = None
        protocol_uris = tuple()

        award_manual = None

    return pipes.PipelineEnd(dataset_path, lifters, context)
    return pipes.SDSPipeline(dataset_path, lifters, context)  # shouldn't need network


def main(path=PathL.cwd(), time_now=None,
         export_local=False, export_parent_path=None,
         _entry_point=False, validate=False, **kwargs):

    # ('../resources/DatasetTemplate')
    pipeline = makeValidator(path, time_now=time_now)
    data = pipeline.data

    if _entry_point:
        from sparcur.simple.export import export_blob
        export_blob_path = export_blob(
            data, 'curation-export.json', time_now=time_now,
            export_parent_path=export_parent_path if export_parent_path is not None else path,
            **kwargs)
        return export_blob_path
    else:
        return data


if __name__ == '__main__':
    from sparcur.simple.utils import pipe_main
    pipe_main(main)
# validate.py ends here



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::*Fetch][Fetch:1]]
from sparcur.simple.fetch_metadata_files import main as files
from sparcur.simple.fetch_remote_metadata import main as rmeta


def main(path=Path.cwd(), **kwargs):
    if path is None or not path.find_cache_root() in (path, *path.parents):
        from sparcur.simple.pull import main as pull
        path = pull(path=path, n_jobs=n_jobs, **kwargs)

    # FIXME these can be run in parallel
    # python is not its own best glue code ...
    rmeta(path=path)
    files(path=path)
    return path


if __name__ == '__main__':
    from sparcur.simple.utils import pipe_main
    pipe_main(main, after=print)
# Fetch:1 ends here



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::dataset-transform][dataset-transform]]
from pathlib import Path
from sparcur import schemas as sc
from sparcur import pipelines as pipes
from sparcur.core import DictTransformer as DT


def __apply_step(step, spec, data, **kwargs):
    return step(data, spec, **kwargs)


def popl(data, pops, source_key_optional=False):
    popped = list(DT.pop(data, pops, source_key_optional))
    return data


def simple_add(data, adds):
    pass


def execute_pipeline(pipeline, data):
    for func, *args, kwargs in pipeline:
        # man variable arity is a pain to deal with here
        # where are lambda lists when you need them :/
        # FIXME maybe we can make steps functional instead of mutating?
        if not kwargs:
            kwargs = {}

        func(data, *args, **kwargs)

    return data


def __wd(transformer):  # not needed atm since transformers do in place modification
    def inner(data, *args, **kwargs):
        transformer(data, *args, **kwargs)
        return data


def schema_validate(data, schema, fail=False, pipeline_stage_name=None):
    if isinstance(schema, type):
        # we were passed a class so init it
        # doing it this way makes it easier to
        # use remote schemas that hit the network
        # since the network call doesn't have to
        # happen at the top level but can mutate
        # the class later before we init here
        schema = schema()

    ok, norm_or_error, data = schema.validate(data)
    if not ok:
        if fail:
            logd.error('schema validation has failed and fail=True')
            raise norm_or_error

        if 'errors' not in data:
            data['errors'] = []

        if pipeline_stage_name is None:
            pipeline_stage_name = f'Unknown.checked_by.{schema.__class__.__name__}'

        data['errors'] += norm_or_error.json(pipeline_stage_name)
        # TODO make sure the step is noted even if the schema is the same


simple_moves = (
    [['structure', 'dirs',], ['meta', 'dirs']],  # FIXME not quite right ...
    [['structure', 'files',], ['meta', 'files']],
    [['structure', 'size',], ['meta', 'size']],
    [['remote_dataset_metadata'], ['inputs', 'remote_dataset_metadata']],
    *pipes.SDSPipeline.moves[3:]
)

# function, *args, **kwargs
# functions should accept data as the first argument
core_pipeline = (
    # FIXME validation of dataset_raw is not being done right now
    (DT.copy, pipes.SDSPipeline.copies, dict(source_key_optional=True)),
    (DT.move, simple_moves, dict(source_key_optional=True)),
    # TODO clean has custom behavior
    (popl, pipes.SDSPipeline.cleans, dict(source_key_optional=True)),
    (DT.derive, pipes.SDSPipeline.derives, dict(source_key_optional=True)),
    #(DT.add, pipes.SDSPipeline.adds),  # FIXME lifter issues
    (schema_validate, sc.DatasetOutSchema, None),
    # extras (missing)
    # xml files
    # contributors
    # submission
    (DT.copy, pipes.PipelineExtras.copies, dict(source_key_optional=True)),
    # TODO clean has custom behavior
    (DT.update, pipes.PipelineExtras.updates, dict(source_key_optional=True)),
    (DT.derive, pipes.PipelineExtras.derives, dict(source_key_optional=True)),
    #(DT.add, pipes.PipelineExtras.adds),  # TODO and lots of evil custom behavior here
    # TODO filter_failures
    (schema_validate, sc.DatasetOutSchema, None),
    (pipes.PipelineEnd._indexes, None),  # FIXME this is conditional and in adds
    # TODO pipeline end has a bunch of stuff in here
    (schema_validate, sc.PostSchema, dict(fail=True)),
)


def main(path=Path.cwd(), path_dataset_raw=None, dataset_raw=None, **kwargs):
    # FIXME TODO need to follow the behavior of main in extract
    if dataset_raw is None:
        if path_dataset_raw is None:
            cache = path.cache
            if not cache.is_dataset():
                raise TypeError('can only run on a single dataset')
            from sparcur.simple.extract import main as extract
            dataset_raw = extract(path)
        else:
            from sparcur.utils import path_ir
            dataset_raw = path_ir(path_dataset_raw)

    data = execute_pipeline(core_pipeline, dataset_raw)
    breakpoint()


if __name__ == '__main__':
    from sparcur.simple.utils import pipe_main
    pipe_main(main, after=print)
# dataset-transform ends here



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::fetch_metadata_files.py][fetch_metadata_files.py]]
from itertools import chain
from sparcur import exceptions as exc
from sparcur.utils import log, logd
from sparcur.paths import Path
from sparcur.datasets import DatasetStructure
from sparcur.simple.utils import fetch_paths_parallel, rglob

# fetch metadata files
fetch_prefixes = (
    ('dataset_description', 'glob'),
    ('subjects',            'glob'),
    ('samples',             'glob'),
    ('submission',          'glob'),
    ('manifest',           'rglob'),  # XXX NOTE the rglob here
)


def _from_path_fetch_metadata_files_simple(path, fetch=True):
    """ transitive yield paths to all metadata files, fetch them from
        the remote if fetch == True """
    for glob_prefix, glob_type in fetch_prefixes:
        if glob_type == 'rglob':
            gp0 = glob_prefix[0]
            pattern = f'[{gp0.upper()}{gp0}]{glob_prefix[1:]}*'
            yield from rglob(path, pattern)
            continue
        ds = DatasetStructure(path)
        for path_to_metadata in ds._abstracted_paths(glob_prefix,
                                                     glob_type=glob_type,
                                                     fetch=fetch):  # FIXME fetch here is broken
            yield path_to_metadata


def _from_path_fetch_metadata_files_parallel(path, n_jobs=12):
    """ Fetch all metadata files within the current path in parallel. """
    paths_to_fetch = _from_path_fetch_metadata_files_simple(path, fetch=False)
    try:
        first = next(paths_to_fetch)
        paths_to_fetch = chain((first,), paths_to_fetch)
    except StopIteration:
        log.warning('No paths to fetch, did you pull the file system metadata?')
        return

    # FIXME passing in a generator here fundamentally limits the whole fetching
    # process to a single thread because the generator is stuck feeding from a
    # single process, IF you materialize the paths first then the parallel fetch
    # can actually run in parallel, but bugs/errors encountered late in collecting
    # the paths will force all previous work to be redone
    # XXX as a result of this we use the posix find command to implement rglob
    # in a way that is orders of magnitude faster
    paths_to_fetch = list(paths_to_fetch)
    fetch_paths_parallel(paths_to_fetch, n_jobs=n_jobs)


def from_path_fetch_metadata_files(path, n_jobs=12):
    """ fetch metadata files located within a path """
    #if n_jobs <= 1:
        #gen = _from_path_fetch_metadata_files_simple(path)
        # FIXME broken ??? somehow abstracted paths doesn't fetch when
        # we run in directly, or somehow fetch_paths_parallel does something
        # different
        #paths = list(gen)
    #else:
    _from_path_fetch_metadata_files_parallel(path, n_jobs=n_jobs)


def main(path=Path.cwd(), n_jobs=12, **kwargs):
    if path is None or path.find_cache_root() not in (path, *path.parents):
        from sparcur.simple.pull import main as pull
        path = pull(path=path, n_jobs=n_jobs, **kwargs)

    from_path_fetch_metadata_files(path, n_jobs=n_jobs)
    return path


if __name__ == '__main__':
    from sparcur.simple.utils import pipe_main
    pipe_main(main, after=print)
# fetch_metadata_files.py ends here



#!/usr/bin/env python3
from sparcur.config import auth
__doc__ = f"""
SPARC curation cli for fetching, validating datasets, and reporting.
Usage:
    spc configure
    spc clone    [options] <project-id>
    spc pull     [options] [<directory>...]
    spc refresh  [options] [<path>...]
    spc fetch    [options] [<path>...]
    spc find     [options] --name=<PAT>...
    spc status   [options]
    spc meta     [options] [<path>...]
    spc rmeta    [options]
    spc export   [schemas protcur protocols] [options] [<path>...]
    spc report   all     [options]
    spc report   size    [options] [<path>...]
    spc report   tofetch [options] [<directory>...]
    spc report   terms   [anatomy cells subcelluar] [options]
    spc report   terms   [hubmap hubmap-anatomy]    [options]
    spc report   overview [<path>...]               [options]
    spc report   anno-tags <tag>...                 [options]
    spc report   [samples-values subjects-values]   [options]
    spc report   [access filetypes pathids errors]  [options]
    spc report   [completeness keywords subjects]   [options]
    spc report   [contributors samples milestones]  [options]
    spc report   [protocols changes test mbf mis]   [options]
    spc shell    [affil integration protocols exit] [options]
    spc shell    [(dates [<path>...]) sheets]       [options]
    spc server   [options]
    spc apinat   [options] <path-in> <path-out>
    spc tables   [options] [<directory>...]
    spc annos    [options] [fetch export shell]
    spc feedback <feedback-file> <feedback>...
    spc missing  [options]
    spc xattrs   [options]
    spc goto     <remote-id>
    spc fix      [options] [duplicates mismatch cache] [<path>...]
    spc fix      [options] [bf-to-pn] [<path>...]
    spc stash    [options --restore] <path>...
    spc make-url [options] [<id-or-path>...]
    spc show     [schemas rmeta (export [json ttl])]  [options] [<project-id>]
    spc show     protcur  [json ttl]                  [options]
    spc sheets   [update cache] [options] <sheet-name>
    spc fab      [meta] [options]

Commands:
    configure   run commands to check and get auth creditials
    clone       clone a remote project (creates a new folder in the current directory)

    pull        retrieve remote file structure

                options: --empty
                       : --sparse-limit

    refresh     retrieve remote file sizes and fild ids (can also fetch using the new data)

                options: --fetch
                       : --level
                       : --only-no-file-id

    fetch       fetch remote data based on local metadata (NOTE does NOT refresh first)

                options: --level
                       : --mbf      fetch mbf xml metadata and only for specific datasets

    find        list unfetched files with option to fetch

                options: --name=<PAT>...  glob options should be quoted to avoid expansion
                       : --existing       include existing files in search
                       : --refresh        refresh matching files
                       : --fetch          fetch matching files
                       : --level

    status      list existing files where local meta does not match cached

    meta        display the metadata the current folder or specified paths

                options: --diff     diff the local and cached metadata
                       : --uri      render the uri for the remote
                       : --browser  navigate to the human uri for this file
                       : --human
                       : --context  include context, e.g. dataset

    rmeta       retrieve metadata about files/folders from the remote

    export      export extracted data to json (and everything else)

                schemas         export schemas from python to json

                options: --latest   run derived pipelines from latest json
                       : --partial  run derived pipelines from the latest partial json export
                       : --open=P   open the output file with specified command
                       : --show     open the output file using xopen
                       : --mbf      extract and export mbf embedded metadata

    report      generate reports

                all             generate all reports (use with --to-sheets)
                size            dataset sizes and file counts
                completeness    submission and curation completeness
                filetypes       filetypes used across datasets
                pathids         mapping from local path to cached id
                keywords        keywords used per dataset
                terms           all ontology terms used in the export

                                anatomy
                                cells
                                subcelluar
                                hubmap
                                hubmap-anatomy

                subjects        all headings from subjects files
                samples         all headings from samples files
                contributors    report on all contributors
                errors          list of all errors per dataset
                test            do as little as possible (use with --profile)
                mbf             mbf term report (can use with --unique)
                anno-tags       list anno exact for a curation tag
                protocols       general report on status of protocols
                changes         diff two curation exports
                mis             list summary predicates used per dataset
                access          report on dataset access master vs pipelines
                overview        general dataset information
                samples-values  report all cell values for samples sheets
                subjects-values report all cell values for subjects sheets

                options: --raw  run reports on live data without export
                       : --tab-table
                       : --to-sheets
                       : --sort-count-desc
                       : --unique
                       : --uri
                       : --uri-api
                       : --uri-html
                       : --debug
                       : --export-file=PATH
                       : --protcur-file=PATH
                       : --ttl-file=PATHoURI
                       : --ttl-compare=PATHoURI
                       : --published

    show        show an export file

                schemas         show the latest schema export folder
                rmeta           show the rmeta cache folder
                export          show the latest project level export

                                json
                                ttl

                protcur         show the latest protcur export

                                json
                                ttl

                options: --open=P   open the output file with specified command

    shell       drop into an ipython shell

                integration     integration subshell with different defaults
                exit            (use with --profile)

    server      reporting server

                options: --raw  run server on live data without export

    apinat      convert ApiNATOMY json to rdf and serialize to ttl

    missing     find and fix missing metadata
    xattrs      populate metastore / backup xattrs
    goto        given an id cd to the containing directory
                invoke as `pushd $(spc goto <id>)`
    dedupe      find and resolve cases with multiple ids
    fix         broke something? put the code to fix it here

                mismatch
                duplicates
    stash       stash a copy of the specific files and their parents
    make-url    return urls for blackfynn dataset ids, or paths
    fab         fabricate something

                meta        make fake metadata for a locally updated file

Options:
    -f --fetch              fetch matching files
    -R --refresh            refresh matching files
    -r --rate=HZ            sometimes we can go too fast when fetching [default: 5]
    -l --limit=SIZE_MB      the maximum size to download in megabytes [default: 2]
                            use zero or negative numbers to indicate no limit
    -L --level=LEVEL        how deep to go in a refresh
                            used by any command that acceps <path>...
    -p --pretend            if the defult is to act, dont, opposite of fetch

    -h --human              print human readable values
    -b --browser            open the uri in default browser
    -u --uri                print the human uri for the path in question
    -a --uri-api            print the api uri for the path in question
       --uri-html           print the html uri for the path in question
    -c --context            include context for a file e.g. dataset
    -n --name=<PAT>         filename pattern to match (like find -name)
    -e --empty              only pull empty directories
    -x --exists             when searching include files that have already been pulled
    -m --only-meta          only pull known dataset metadata files
    -z --only-no-file-id    only pull files missing file_id
    -o --overwrite          fetch even if the file exists
    --project-path=<PTH>    set the project path manually
    --sparse-limit=COUNT    package count that forces a sparse pull [default: {auth.get('sparse-limit')}]
                            use zero or negative numbers to indicate no limit

    -F --export-file=PATH   run reports on a specific export file
    -t --tab-table          print simple table using tabs for copying
    -A --latest             run derived pipelines from latest json
    -P --partial            run derived pipelines from the latest partial json export
    -W --raw                run reporting on live data without export
    --published             run on the latest published export
    --to-sheets             push report to google sheets
    --protcur-file=PATH     location of protcur jsonld file
    --ttl-file=PATHoURI     location of ttl file (uses latest if not specified)
    --ttl-compare=PATHoURI  location of ttl file for comparison
    --preview               run export and reporting in preview mode
                            if not set auth.get('preview') takes priority

    -S --sort-size-desc     sort by file size, largest first
    -C --sort-count-desc    sort by count, largest first

    -O --open=PROGRAM       open the output file with program
    -w --show               open the output file
    -U --upload             update remote target (e.g. a google sheet) if one exists
    -N --no-google          hack for ipv6 issues
    -D --diff               diff local vs cache
    --force                 force the regeneration of a cached file

    --port=PORT             server port [default: 7250]

    -j --jobs=N             number of jobs to run             [default: 12]
    -d --debug              drop into a shell after running a step
    -v --verbose            print extra information
    --profile               profile startup performance
    --local                 ignore network issues
    --no-network            do not make any network requests (incomplete impl)
    --mbf                   fetch/export mbf related metadata
    --unique                return a unique set of values without additional info

    --log-level=LEVEL       set python logging log level
    --log-path=PATH         folder where logs are saved       [default: {auth.get_path('log-path')}]
    --cache-path=PATH       folder where remote data is saved [default: {auth.get_path('cache-path')}]
    --export-path=PATH      base folder for exports           [default: {auth.get_path('export-path')}]

    --project-id=PID        alternate way to pass project id  [default: {auth.get('remote-organization')}]

    --hypothesis-group-name=NAME  hypothesis group name for protcur  [default: {auth.get('hypothesis-group-name')}]
    --hypothesis-cache-file=PATH  path to hyputils json cache file
    --i-know-what-i-am-doing      don't use this unless you already know what it does
"""

from time import time
start = time()

import os
import sys
import json
import errno
import types
from itertools import chain
from collections import Counter, defaultdict
import idlib
import htmlfn as hfn
import ontquery as oq
start_middle = time()
import augpathlib as aug
from pyontutils import clifun as clif
from pyontutils.core import OntResGit
from pyontutils.utils import UTCNOWISO, subclasses
from pyontutils.config import auth as pauth
from terminaltables import AsciiTable

from sparcur import reports  # top level
from sparcur import datasets as dat
from sparcur import exceptions as exc
from sparcur.core import JT
from sparcur.core import OntId, OntTerm, adops
from sparcur.utils import GetTimeNow  # top level
from sparcur.utils import log, logd, loge, bind_file_handler
from sparcur.utils import register_type, fromJson
from sparcur.paths import Path, StashPath
from sparcur.state import State
from sparcur.protocols import ProtocolData

try:
    breakpoint
except NameError:
    from IPython import embed as breakpoint

stop = time()


class Options(clif.Options):

    @property
    def export_schemas_path(self):
        return Path(self.export_path) / 'schemas'  # FIXME not sure if correct

    @property
    def export_protcur_base(self):
        return Path(self.export_path) / 'protcur'  # FIXME not sure if correct

    @property
    def export_protocols_base(self):
        return Path(self.export_path) / 'protocols'  # FIXME not sure if correct

    @property
    def project_id(self):
        return self._args['<project-id>'] or self._args['--project-id']

    @property
    def project_path(self):
        pp = self._args['--project-path']
        if pp:
            return Path(pp).expanduser().resolve()

    @property
    def export_file(self):
        ef = self._args['--export-file']
        if ef:
            return Path(ef).expanduser().resolve()

    @property
    def protcur_file(self):
        pf = self._args['--protcur-file']
        if pf:
            return Path(pf).expanduser().resolve()

    @property
    def jobs(self):
        return int(self._args['--jobs'])

    @property
    def limit(self):
        l = int(self._args['--limit'])
        if l >= 0:
            return l

    @property
    def sparse_limit(self):
        return int(self._args['--sparse-limit'])

    @property
    def level(self):
        return int(self._args['--level']) if self._args['--level'] else None

    @property
    def rate(self):
        return int(self._args['--rate']) if self._args['--rate'] else None

    @property
    def fetch(self):
        return self._args['--fetch'] or self._default_fetch

    @property
    def mbf(self):
        # deal with the fact that both mbf and --mbf are present
        return self._args['--mbf'] or self._default_mbf

    @property
    def show(self):
        # deal with the fact that both show and --show are present
        return self._args['--show'] or self._default_show

    @property
    def preview(self):
        # FIXME this should also be determined by
        # checking the schema version of the data being run
        return self._args['--preview'] or auth.get('preview')

    @property
    def log_level(self):
        ll = self._args['--log-level']
        if ll is None:
            return
        if ll.isdigit() or ll[0] == '-' and ll[1:].isdigit():
            return int(ll)
        else:
            return ll


class Dispatcher(clif.Dispatcher):
    spcignore = ('.git',
                 '.~lock',)

    def _export(self, export_class, export_source_path=None, org_id=None):

        if export_source_path is None:
            export_source_path = self.cwd

        export_source_path = self.cwd  # XXX note the override
        export = export_class(
            self.options.export_path,
            export_source_path,
            self._folder_timestamp,
            self._timestamp,
            self.options.latest,
            self.options.partial,
            (self.options.open
             if self.options.open else
             self.options.show),
            org_id,
            self.options.export_protcur_base,
            no_network=(self.options.no_network or self.options.no_google),
        )
        return export

    def _print_table(self, rows, title=None, align=None, ext=None):
        """ ext is only used when self.options.server -> True """
        def asStr(c):
            if hasattr(c, 'asStr'):
                return c.asStr()
            else:
                return str(c)

        def asAsciiRows(rows):
            return [[asStr(cell) for cell in row] for row in rows]

        def simple_tsv(rows):
            return '\n'.join('\t'.join(row) for row in asAsciiRows(rows)) + '\n'

        if self.options.tab_table:
            if title:
                print(title)

            print(simple_tsv(rows))

        elif self.options.server:
            if ext is not None:
                if ext == '.tsv':
                    nowish = UTCNOWISO('seconds')
                    fn = json.dumps(f'{title} {nowish}')
                    return (
                        simple_tsv(rows), 200,
                        {'Content-Type': 'text/tsv; charset=utf-8',
                         'Content-Disposition': f'attachment; filename={fn}'})
                if isinstance(ext, types.FunctionType):
                    return ext(hfn.render_table(rows[1:], *rows[0]), title=title)
                else:
                    return 'Not found', 404

            return hfn.render_table(rows[1:], *rows[0]), title
        else:
            table = AsciiTable(asAsciiRows(rows), title=title)
            if align:
                assert len(align) == len(rows[0])
                table.justify_columns = {i:('left' if v == 'l'
                                            else ('center' if v == 'c'
                                                  else ('right'
                                                        if v == 'r' else
                                                        'left')))
                                         for i, v in enumerate(align)}
            print(table.table)
            if ext:
                return rows, title

    def _print_paths(self, paths, title=None):
        if self.options.sort_size_desc:
            key = lambda ps: -ps[-1]
        else:
            key = lambda ps: ps

        def derp(p):
            if p.cache is None:
                raise exc.NoCachedMetadataError(p)

        rows = [['Path', 'size', '?'],
                *((p, s.hr
                   if isinstance(s, aug.FileSize) else
                   s, 'x' if p.exists() else '')
                  for p, s in
                  sorted(([p, ('/' if p.is_dir() else
                               (p.cache.meta.size
                                if p.cache.meta.size else
                                '??')
                               if p.cache.meta else '_')]
                          for p in paths if not derp(p)), key=key))]
        self._print_table(rows, title)


class Main(Dispatcher):
    child_port_attrs = ('anchor',
                        'project_path',
                        'project_id',
                        'bfl',
                        'Remote',
                        '_remote_class',
                        '_cache_class',
                        'Integrator',
                        'summary',
                        'cwd',
                        'cwdintr',
                        '_datasets_with_extension',
                        '_timestamp',
                        '_folder_timestamp',
                        '_pyru_loaded',
                        '_data_ir',
                        '_protcur',
                        )

    # any attr forced on children must be set before super().__init__ is called
    # set timestamp early so that the loggers can use it
    # I don't think this will cause too much trouble ...
    #_timestamp = None
    #_folder_timestamp = None

    # things all children should have
    # kind of like a non optional provides you WILL have these in your namespace
    def __init__(self, options, time_now=GetTimeNow()):
        self._time_now = time_now
        self._timestamp = self._time_now.START_TIMESTAMP
        self._folder_timestamp = self._time_now.START_TIMESTAMP_LOCAL_FRIENDLY
        super().__init__(options)
        if not self.options.verbose:
            log.setLevel('INFO')
            logd.setLevel('INFO')
            loge.setLevel('INFO')

        if self.options.log_level:
            log.setLevel(self.options.log_level)
            logd.setLevel(self.options.log_level)
            loge.setLevel(self.options.log_level)

        if self.options.configure:
            return

        if self.options.project_path:
            self.cwd = Path(self.options.project_path).resolve()
        else:
            self.cwd = Path.cwd()

        # bind paths
        from sparcur.backends import PennsieveRemote
        from sparcur.paths import PennsieveCache
        self._cache_class = PennsieveCache
        self._remote_class = PennsieveRemote

        # setup integrator
        from sparcur.curation import Integrator
        self.Integrator = Integrator
        self.Integrator.rate = self.options.rate
        self.Integrator.no_google = self.options.no_google

        self.cwdintr = self.Integrator(self.cwd)

        # deal with no network case
        if self.options.no_network:
            from sparcur.sheets import Sheet
            Sheet._only_cache = True

        # pass debug along (sigh)
        from augpathlib import RemotePath, AugmentedPath  # for debug
        AugmentedPath._debug = self.options.debug
        RemotePath._debug = self.options.debug

        # FIXME populate this via decorator
        if (self.options.export and not
            (self.options.schemas or self.options.protcur or self.options.protocols) or
            (self.options.report and self.options.raw)):
            pass  # simplify setup for consumers beyond this file
        elif (self.options.clone or
              self.options.meta or
              (self.options.show and not self.options.export) or
              self.options.sheets or
              self.options.goto or
              self.options.server or
              self.options.apinat or
              self.options.tofetch or  # size does need a remote but could do it lazily
              self.options.filetypes or
              self.options.anno_tags or
              self.options.status or  # eventually this should be able to query whether there is new data since the last check
              self.options.pretend or
              self.options.annos or
              self.options.fab or
              (self.options.fix and self.options.cache) or
              (self.options.report and self.options.size) or
              (self.options.report and self.options.export_file) or
              (self.options.report and self.options.protocols) or
              (self.options.export and self.options.schemas) or
              (self.options.export and self.options.protocols) or  # FIXME if protocols require export ...
              (self.options.export and self.options.protcur) or  # FIXME if protocols require export ...
              (self.options.find and not (self.options.fetch or self.options.refresh))):
            # short circuit since we don't know where we are yet
            self.Integrator.no_google = True
            if self.options.clone:
                self._setup_bfl()
            return

        elif (self.options.pull or
              self.options.mismatch or
              self.options.stash or
              self.options.contributors or
              self.options.make_url or
              self.options.find or
              self.options.missing):
            self.Integrator.no_google = True

        self._setup_local()  # if this isn't run up here the internal state of the program get's wonky

        if (self.options.report and
            not self.options.raw and
            not self.options.access):
            self.Integrator.setup(local_only=True)  # FIXME sigh
        else:
            try:
                self._setup_bfl()
            except BaseException as e:
                if self.options.local:
                    from sparcur.pennsieve_api import FakeBFLocal
                    log.exception(e)
                    self.Remote._api = FakeBFLocal(self.anchor.id, self.anchor)
                    self.Remote.anchorTo(self.anchor)
                else:
                    raise e

        if self.options.export or self.options.shell or self.options.raw:
            self._setup_export()
            #self._setup_ontquery()

    def _setup_local(self):
        self.Remote = self._cache_class._remote_class
        self.Remote._async_rate = self.options.rate

        local = self.cwd

        # we have to start from the cache class so that
        # we can configure
        try:
            _cpath = local.cache  # FIXME project vs subfolder
            if _cpath is None:
                raise exc.NoCachedMetadataError  # FIXME somehow we decided not to raise this!??!

            self.anchor = _cpath.anchor
        except exc.NoCachedMetadataError as e:
            root = local.find_cache_root()
            if root is not None:
                self.anchor = root.cache
                if local.skip_cache:
                    print(f'{local} is ignored!')
                    sys.exit(112)

                raise NotImplementedError('TODO recover meta?')
            else:
                print(f'{local} is not in a project!')
                sys.exit(111)

        # replace a bottom up anchoring rule with top down
        if self.options.i_know_what_i_am_doing:
            pass
        else:
            self.anchor.anchorClassHere(remote_init=False)

        self.project_path = self.anchor.local
        from sparcur.curation import Summary
        self.summary = Summary(self.project_path)
        Summary._n_jobs = self.options.jobs
        if self.options.debug:
            Summary._debug = True

    def _setup_bfl(self):
        if self.options.clone:
            # special case where we may not have all the info
            # needed to get self.anchor
            project_id = self.options.project_id
            if project_id is None:
                print('no remote project id listed')
                sys.exit(4)
            # given that we are cloning it makes sense to _not_ catch a connection error here
            # FIXME putting this here breaks testing because it hops outside
            # the usual process for creating a new remote class in _setup_bfl
            self.Remote = self._remote_class._new(Path, self._cache_class)
            try:
                self.Remote.init(project_id)
            except exc.MissingSecretError:
                print(f'missing api secret entry for {project_id}')
                sys.exit(11)

            return

        if not self.options.i_know_what_i_am_doing:
            self.Remote._setup()
            self.Remote.init(self.anchor.id)

        self.bfl = self.Remote._api
        State.bind_blackfynn(self.bfl)

    def _setup_export(self):
        self.Integrator.setup()

    def _setup_ontquery(self):
        # this should all be in SciGraph by now
        return
        # FIXME this should be in its own setup method
        # pull in additional graphs for query that aren't loaded properly
        RDFL = oq.plugin.get('rdflib')
        olr = Path(pauth.get_path('ontology-local-repo'))
        branch = 'origin/methods'
        # FIXME the blast radius on this thing is of unknown size :/
        for fn in ('methods', 'methods-helper', 'methods-core'):
            org = OntResGit(olr / f'ttl/{fn}.ttl', ref=branch)
            OntTerm.query.ladd(RDFL(org.graph, OntId))

    _pyru_loaded = False

    def _data_ir(self, org_id=None):  # FIXME org_id should NOT implicitly indicate that we are in no_export
        if hasattr(self, '_cache_data_ir'):
            return self._cache_data_ir

        if self.options.export_file or self.options.published:
            if not self.__class__._pyru_loaded:
                self.__class__._pyru_loaded = True
                from pysercomb.pyr import units as pyru
                iso8601s = (pyru.Iso8601DurationTime,
                            pyru.Iso8601DurationDatetime,
                            pyru.Iso8601DurationDate,)
                [register_type(c, c.tag) for c in
                (pyru._Quant, pyru.Range, pyru.Approximately, *iso8601s)]
                pyru.Term._OntTerm = OntTerm  # the tangled web grows ever deeper :x

        if self.options.raw:
            data = self.summary.data_for_export(UTCNOWISO())

        elif self.options.export_file:
            with open(self.options.export_file, 'rt') as f:
                data = fromJson(json.load(f))
        elif self.options.published:
            url = auth.get('export-url')
            if url is not None:
                import requests
                raw = requests.get(url).json()
                data = fromJson(raw)
            else:
                raise ValueError('must set export-url')
        else:
            from sparcur import export as ex
            data = self._export(ex.Export, org_id=org_id).latest_ir

        self._cache_data_ir = data
        return data

    def _protcur(self):
        #from sparcur import export as ex
        #export = self._export(ex.Export, org_id=self.options.project_id)  # FIXME org_id not needed ...
        if self.options.raw:  # FIXME currently lacking impl of LATEST for protcur export
            if not self.options.protcur:
                self.options._args['protcur'] = True  # FIXME hack

            blob_protcur = self.parent.export()

        else:
            # not sure if we are expanding identifiers and such right now ...
            if False and not self.__class__._pyru_loaded:  # FIXME do we need this ?
                self.__class__._pyru_loaded = True
                from pysercomb.pyr import units as pyru
                iso8601s = (pyru.Iso8601DurationTime,
                            pyru.Iso8601DurationDatetime,
                            pyru.Iso8601DurationDate,)
                [register_type(c, c.tag) for c in
                (pyru._Quant, pyru.Range, pyru.Approximately, *iso8601s)]
                pyru.Term._OntTerm = OntTerm  # the tangled web grows ever deeper :x

            if self.options.protcur_file:
                with open(self.options.protcur_file, 'rt') as f:
                    blob_protcur = json.load(f)
                    #blob_protcur = fromJson(json.load(f))  # FIXME do we need to run fromJson on this?
                    # way too many uri lookups and validations are done by fromJson

            else:
                latest_path = self.options.export_protcur_base / 'LATEST'
                latest_partial_path = self.options.export_protcur_base / 'LATEST_PARTIAL'
                path = (latest_partial_path if self.options.partial else latest_path) / 'protcur.json'
                with open(path, 'rt') as f:
                    blob_protcur = json.load(f)

        return blob_protcur

    @property
    def project_name(self):
        return self.anchor.name
        #return self.bfl.organization.name

    @property
    def project_id(self):
        #self.bfl.organization.id
        return self.anchor.id

    @property
    def datasets(self):
        # XXX DO NOT YIELD DIRECTLY FROM self.anchor.children
        # unless you are cloning or something like that
        # because cache.children completely ignores existing
        # files and folders and there really isn't a safe way
        # to use it once files already exist because then
        # viewing the cached children would move all the folders
        # around, file under sigh, yes fix CachePath construction
        for local in self.datasets_local:
            yield local.cache

        #yield from self.anchor.children  # NOT OK TO YIELD FROM THIS URG

    @property
    def datasets_remote(self):
        for d in self.anchor.remote.children:
            # FIXME lo the crossover (good for testing assumptions ...)
            #yield d.local
            yield d

    @property
    def datasets_local(self):
        for d in self.anchor.local.children: #self.datasets:
            if d.exists():
                yield d

    ###
    ## vars
    ###

    @property
    def directories(self):
        return [Path(string_path).absolute()
                for string_path in self.options.directory]

    @property
    def paths(self):
        return [Path(string_path).absolute()
                for string_path in self.options.path]

    @property
    def _paths(self):
        """ all relevant paths determined by the flags that have been set """
        # but if you use the generator version of _paths
        # then if you add a folder to the previous path
        # then it will yeild that folder! which is SUPER COOL
        # but breaks lots of asusmptions elsehwere
        paths = self.paths
        if not paths:
            paths = self.cwd,  # don't call Path.cwd() because it may have been set from --project-path

        if self.options.only_meta:
            paths = (mp.absolute()
                     for p in paths
                     for mp in dat.DatasetStructureLax(p).meta_paths)
            yield from paths
            return

        yield from self._build_paths(paths)

    def _build_paths(self, paths):
        def inner(paths, level=0, stop=self.options.level):
            """ depth first traversal of children """
            for path in paths:
                if self.options.only_no_file_id:
                    if (path.is_broken_symlink() and
                        (path.cache.meta.file_id is None)):
                        yield path
                        continue

                elif self.options.empty:
                    if path.is_dir():
                        try:
                            next(path.children)
                            # if a path has children we still want to
                            # for empties in them to the level specified
                        except StopIteration:
                            yield path
                    else:
                        continue
                else:
                    yield path

                if stop is None:
                    if self.options.only_no_file_id:
                        for rc in path.rchildren:
                            if (rc.is_broken_symlink() and
                                rc.cache.meta.file_id is None):
                                yield rc
                    else:
                        yield from path.rchildren

                elif level <= stop:
                    yield from inner(path.children, level + 1)

        yield from inner(paths)

    @property
    def _dirs(self):
        for p in self._paths:
            if p.is_dir():
                yield p

    @property
    def _not_dirs(self):
        for p in self._paths:
            if not p.is_dir():
                yield p

    def configure(self):
        self._configure(self.options.hypothesis_group_name)

    @staticmethod
    def _configure(hypothesis_group_name=None):
        # the order here is very precise so that the absolutely
        # required and easiest to get right creds go first

        # protocols io
        print('protocols io starting')
        idlib.Pio._setup()
        print('protocols io ok')

        # google sheets readonly
        print('google sheets readonly starting')
        from sparcur.sheets import Organs
        oro = Organs(readonly=True)
        oro.fetch()
        # FIXME googapis I think
        print('google sheets readonly ok')

        # scigraph
        print('scigraph starting')
        print('scigraph api', pauth.get('scigraph-api'))
        next(OntTerm.query(label='brain', prefix='UBERON'))
        print('scigraph ok')

        # pennsieve
        print('pennsieve starting')
        from sparcur.simple.utils import backend_pennsieve
        project_id = auth.get('remote-organization')
        print('pennsieve project_id', project_id)
        PennsieveRemote = backend_pennsieve(project_id)
        root = PennsieveRemote(project_id)
        datasets = list(root.children)
        print('pennsieve ok')

        # google sheets readwrite
        print('google sheets readwrite starting')
        orw = Organs(readonly=False)
        orw.fetch()
        print('google sheets readwrite ok')

        # hypothesis
        print('hypothes.is starting')
        from hyputils import hypothesis as hyp
        group_name = (
            hypothesis_group_name
            if hypothesis_group_name else
            auth.get('hypothesis-group-name'))
        print('hypothesis group_name', group_name)
        group_id = auth.user_config.secrets('hypothesis', 'group', group_name)
        get_annos = hyp.Memoizer(group=group_id)
        get_annos.api_token = auth.get('hypothesis-api-key')
        values = get_annos.get_annos_from_api(max_results=1)
        assert values, 'no values means api_token or group_id are wrong'
        print('hypothes.is ok')

    def clone(self):
        # make sure that we aren't in a project already
        existing_root = self.cwd.find_cache_root()
        if existing_root is not None and existing_root != self.cwd:
            message = f'fatal: already in project located at {existing_root.resolve()!r}'
            print(message)
            sys.exit(3)

        try:
            if self.options.project_path:
                anchor = self.Remote.anchorTo(self.cwd, create=True)
            else:
                anchor = self.Remote.dropAnchor(self.cwd)
        except exc.DirectoryNotEmptyError:
            message = (f'fatal: destination path already '
                       'exists and is not an empty directory.')
            print(message)
            sys.exit(2)
        except BaseException as e:
            log.exception(e)
            sys.exit(11111)

        anchor.local_data_dir_init()

        self.anchor = anchor
        self.project_path = self.anchor.local
        with anchor:
            # update self.cwd so pull sees the right thing
            self.cwd = Path.cwd()
            self._clone_pull()

    def _clone_pull(self):
        """ wow this really shows that the pull -> bootstrap workflow
            is completely broken and stupidly slow """

        # populate just the dataset folders
        datasets = list(self.anchor.children)
        # populate the dataset metadata
        self.rmeta(use_cache_path=True, exist_ok=True)

        # mark sparse datasets so that we don't have to
        # fiddle with detecting sparseness during bootstrap
        sparse_limit = self.options.sparse_limit
        [d._sparse_materialize(sparse_limit=sparse_limit) for d in datasets]
        sparse = [d for d in datasets if d.is_sparse()]  # sanity check


        skip = auth.get_list('datasets-no')  # FIXME this should be materialized as well

        # pull all the files
        self.pull()

    def pull(self):
        # FIXME this still isn't quite right because it does not correctly
        # deal dataset renaming, but it is much closer
        from joblib import Parallel, delayed
        dirs = self.directories
        dirs = dirs if dirs else (self.cwd,)
        for d in dirs:
            d.pull(
                time_now=self._time_now,
                debug=self.options.debug,
                n_jobs=12,
                log_level='DEBUG' if self.options.verbose else 'INFO',
                Parallel=Parallel,
                delayed=delayed,)

    def refresh(self):
        paths = self.paths
        cwd = self.cwd
        if not paths:
            paths = cwd,

        to_root = sorted(set(parent
                             for path in paths
                             for parent in path.parents
                             if parent.cache is not None),
                         key=lambda p: len(p.parts))

        if self.options.pretend:
            ap = list(chain(to_root, self._paths))
            self._print_paths(ap)
            print(f'total = {len(ap):<10}rate = {self.options.rate}')
            return

        self._print_paths(chain(to_root, self._paths))

        from pyontutils.utils import Async, deferred
        hz = self.options.rate
        fetch = self.options.fetch
        limit = self.options.limit

        drs = [d.remote for d in chain(to_root, self._dirs)]

        if not self.options.debug:
            refreshed = Async(rate=hz)(deferred(r.refresh)(
                update_data_on_cache=r.cache.is_file() and
                r.cache.exists())
                                       for r in drs)
        else:
            refreshed = [r.refresh(update_data_on_cache=r.cache.is_file() and
                                   r.cache.exists()) for r in drs]

        moved = []
        parent_moved = []
        for new, r in zip(refreshed, drs):
            if new is None:
                log.critical('utoh')

            oldl = r.local
            try:
                r.update_cache()  # calling this directly is ok for directories
            except FileNotFoundError as e:
                parent_moved.append(oldl)
                continue
            except OSError as e:
                if e.errno == errno.ENOTEMPTY:
                    log.error(f'{e}')
                    continue
                else:
                    raise e

            newl = r.local
            if oldl != newl:
                moved.append([oldl, newl])

        if moved:
            self._print_table(moved, title='Folders moved')
            for old, new in moved:
                if old == cwd:
                    log.info(f'Changing directory to {new}')
                    new.chdir()

        if parent_moved:
            self._print_paths(parent_moved, title='Parent moved')

        if not self.options.debug:
            refreshed = Async(rate=hz)(deferred(path.cache.refresh)(
                update_data=fetch, size_limit_mb=limit)
                                       for path in self._not_dirs)

        else:
            for path in self._not_dirs:
                path.cache.refresh(update_data=fetch, size_limit_mb=limit)

    def _datasets_with_extension(self, extension):
        """ Hack around the absurd slowness of python's rglob """

        command = fr"""for d in */; do
    find "$d" \( -type l -o -type f \) -name '*.{extension}' \
    -exec getfattr -n user.bf.id --only-values "$d" \; -printf '\n' -quit ;
done"""
        with os.popen(command) as p:
            string = p.read()

        has_extension = string.split('\n')
        datasets = [d for d in self.datasets if d.id in has_extension]
        return datasets

    def _fetch_mbf(self):
        from pyontutils.utils import Async, deferred
        def fetch_mbf_metadata(dataset):
            xmls = dataset.rglob('*.xml')
            Async(rate=5)(deferred(x.fetch)(size_limit_mb=None)
                          for x in xmls if not x.exists())

        for dataset in self._datasets_with_extension('xml'):
            fetch_mbf_metadata(dataset)

    def fetch(self):
        if self.options.mbf:  # FIXME hack -> fetch '*.xml' or something
            self._fetch_mbf()
            return

        paths = [p for p in self._paths if not p.is_dir()]
        self._print_paths(paths)
        if self.options.pretend:
            return

        from pyontutils.utils import Async, deferred
        hz = self.options.rate
        Async(rate=hz)(deferred(path.cache.fetch)(
            size_limit_mb=self.options.limit)
                       for path in paths
                       if not path.exists()
                       # FIXME need a staging area ...
                       # FIXME also, the fact that we sometimes need content_different
                       # means that there may be silent fetch failures
                       or path.content_different())

    def _check_duplicates(self, datasets):
        # NOTE this is an ok sanity check
        # but cannot catch the duplicate dataset issue
        # which actually comes from calling refresh
        # while iterating through the list of dataset paths
        # which can and does cause datasets move due to
        # changes in their folder name
        if not datasets:
            datasets = list(self.datasets_local)

        counts = Counter([d['id'] if isinstance(d, dict) else d.id
                          for d in datasets])
        bads = [(id, c) for id, c in counts.most_common() if c > 1]
        if bads:
            # FIXME sys.exit ?
            raise BaseException(f'duplicate datasets {bads}')

    def _check_exists(self, dataset_paths):
        bads = [p for p in dataset_paths if not p.exists()]
        if bads:
            raise FileNotFoundError(f'The following do not exist!\n{bads}')

    def export(self):
        OntTerm._nofetch = True  # enable network sandbox for all export activity
        # FIXME no_google couples network and behavior, we need a decoupled version now that we have caching
        from sparcur import export as ex  # FIXME very slow to import
        from sparcur import schemas as sc

        # FIXME export should be able to run without needing any external
        # data the fetch step should happen before the export so that
        # network connections don't creep into the process

        # TODO export source path without having to be in that folder

        if self.options.schemas:
            ex.core.export_schemas(self.options.export_schemas_path)

        elif self.options.protcur or self.options.protocols:
            export = self._export(
                ex.Export,
                org_id=self.options.project_id,  # FIXME org_id not needed ...
            )
            # FIXME dump_path shouldn't need to be passed explicitly
            dump_path = (self.options.export_protcur_base /
                         self._folder_timestamp)
            hgn = self.options.hypothesis_group_name  # FIXME when to switch public/secret?
            if hgn is None:
                raise ValueError('Missing hypothesis-group-name!')
            blob_protcur = export.export_protcur(dump_path,
                                                 hgn,
                                                 rerun_protcur_export=self.options.force,
                                                 direct=True)
            # NOTE --latest will also pull from latest protcur export
            # if it exists I think, the issue is with repeated export
            # of the same content ... as usual, which will confuse
            # EVERYTHING if we want the latest protcur export but
            # not the latest dataset export, this will have to be
            # dealt with when we move to the default state being
            # to update datasets independently where there will only
            # be a fully integrated dataset when we do the conversion
            # to ttl or bulk load into foundry etc.
            if self.options.protocols:  # TODO
                protcur_dump_path = dump_path
                dataset_blobs = (self._data_ir()['datasets']
                                 if self.options.export_file else
                                 [])  # could be worse ...
                dump_path = (self.options.export_protocols_base /
                             self._folder_timestamp)
                dump_path.mkdir(parents=True)
                export.export_protocols(dump_path,
                                        dataset_blobs,
                                        blob_protcur)
                # need to be able to run this from arbitrary dataset blobs
                # in a way that is marginally sane
                raise NotImplementedError('dataset blobs inclusion please')
            else:
                log.info('export protcur completed successfully')
                return blob_protcur

        elif self.options.mbf:
            export = self._export(ex.ExportXml)
            dataset_paths = self._datasets_with_extension('xml')
            self._check_duplicates(dataset_paths)
            self._check_exists(dataset_paths)
            blob_ir, *rest = export.export(dataset_paths=dataset_paths,
                                           jobs=self.options.jobs,
                                           debug=self.options.debug)

            return blob_ir

        elif self.cwd.cache.is_dataset():
            export = self._export(ex.Export)
            blob_ir, intr, dump_path, latest_path = export.export_single_dataset()
            return blob_ir, intr, dump_path, latest_path

        else:
            export = self._export(ex.Export)
            dataset_paths = tuple(self.paths)
            self._check_duplicates(dataset_paths)  # NOTE can be empty
            self._check_exists(dataset_paths)
            noexport = (auth.get_list('datasets-noexport') +
                        auth.get_list('datasets-no'))
            blob_ir, *rest = export.export(dataset_paths=dataset_paths,
                                           exclude=noexport)

            sc.SummarySchema().validate_strict(export.latest_export)

        log.info('export completed successfully')

        if self.options.debug:
            breakpoint()

    def annos(self):
        hgn = self.options.hypothesis_group_name
        if hgn is None:
            raise ValueError('Missing hypothesis-group-name!')

        if self.options.fetch:
            from hyputils import hypothesis as hyp
            group_id = auth.user_config.secrets('hypothesis', 'group', hgn)
            cache_file = Path(hyp.group_to_memfile(group_id + 'sparcur'))
            get_annos = hyp.Memoizer(cache_file, group=group_id)
            get_annos.api_token = auth.get('hypothesis-api-key')  # FIXME ?
            annos = get_annos()
            return

        from protcur.analysis import protc
        from sparcur import pipelines as pipes
        pipe = pipes.ProtcurPipeline(hgn)
        _annos, _lsus = pipe.load()

        # using _annos here as we transition to use ptcdoc.Annotation
        # instead of the base hypothesis annotation as the substrate
        annos, idints, pidints = pipe._idints(_annos)

        pidints = None  # repr issues
        # the per doc use case is rather different from the traditional
        # protcur use case because the traditional protcur use case
        # assumes that protocols routinely cross document boundaries
        # which is absolutely the case when curating from the literature
        # from protocols.io not quite so much

        per_pid = {pid:protc.protcurLang([a.id for a in idannos])
                   # FIXME protcur export has been insanely
                   # slow for a long time and continues to be
                   for pid, idannos in idints.items()}

        if self.options.export:
            base = Path('/tmp/protcur-rkt/')
            base.mkdir(exist_ok=True)

            for pid, string in per_pid.items():
                fn = (pid.identifier.suffix
                    if isinstance(pid, idlib.Pio) and pid.identifier.is_int() else
                      (pid.suffix if isinstance(pid, idlib.Pio._id_class) else
                       pid.replace('/', '_').replace(':', '_'))) + '.ptc'
                with open(base / fn, 'wt') as f:
                    f.write(string)

            with open('/tmp/sparc-protcur.rkt', 'wt') as f:
                f.write(protc.parsed())

        """
        if self.options.raw:
            data = self.summary.data()
        else:
            from sparcur import export as ex
            data = self._export(ex.Export).latest_ir

        dataset_blobs = data['datasets']

        from protcur.analysis import protc
        from sparcur.protocols import ProtcurData
        ProtcurData.populate_annos()

        class ProtocolActual(ProtocolData):  # FIXME so ... bad ...
            @property
            def protocol_uris(self, outer_self=self):  # FIXME this needs to be pipelined
                for d in dataset_blobs:
                    try:
                        yield from adops.get(d, ['meta', 'protocol_url_or_doi'])
                    except exc.NoSourcePathError:
                        pass

        pa = ProtocolActual()
        all_blackfynn_uris = set(pa.protocol_uris_resolved)
        all_hypothesis_uris = set(a.uri for a in protc)
        if self.options.shell or self.options.debug:
            p, *rest = self._paths
            f = self.Integrator(p)
            all_annos = [list(protc.byIri(uri))
                         for uri in f.protocol_uris_resolved]
            breakpoint()
        """

    def tables(self):
        """ print summary view of raw metadata tables, possibly per dataset """

        dat.DatasetStructure._refresh_on_missing = False
        dat.SubmissionFile._refresh_on_missing = False
        dat.SubjectsFile._refresh_on_missing = False
        dat.SamplesFile._refresh_on_missing = False
        directories = self.directories
        directories = directories if directories else (self.cwd,)
        for directory in directories:
            self._tables_dir(directory)

    def _tables_dir(self, directory):
        droot = dat.DatasetStructure(directory)
        if droot.cache.is_dataset():
            datasetdatas = droot,
        elif droot.cache.is_organization():
            datasetdatas = droot.children
        else:
            raise TypeError(f'whats a {type(droot)}? {droot}')

        def do_file(d, ft):
            if hasattr(d, ft):
                f = getattr(d, ft)
                fs = f if isinstance(f, list) else [f]
                for f in fs:
                    if hasattr(f, 'object'):
                        yield (f.object._t())
                    else:
                        message = f'object missing on {ft} for {d}'
                        log.warning(message)
                        yield message

            else:
                yield f'{ft} missing on {d}'

        tables = [t for d in datasetdatas for t in
                  (*do_file(d, 'dataset_description'),
                   *do_file(d, 'submission'),
                   *do_file(d, 'subjects'),
                   *do_file(d, 'samples'),
                   *do_file(d, 'manifest'),
                   ('\n--------------------------------------------------\n'
                      '=================================================='
                    '\n--------------------------------------------------\n'))]

        def safe_repr(t):
            try:
                return repr(t)
            except Exception as e:
                log.exception(e)

        [print(t if isinstance(t, str) else safe_repr(t)) for t in tables]  # FIXME _print_tables?

    def find(self):
        _paths = set()  # set to avoid duplicate paths breaking fetch?
        if self.options.name:  # has to always be true now
            patterns = self.options.name
            path = self.cwd
            for pattern in patterns:
                # TODO filesize mismatches on non-fake
                # no longer needed due to switching to symlinks
                #if '.fake' not in pattern and not self.options.overwrite:
                    #pattern = pattern + '.fake*'

                for file in path.rglob(pattern):
                    _paths.add(file)

        paths = sorted(_paths)
        if paths:
            n_skipped = 0
            paths = [p for p in paths if not p.is_dir()]
            search_exists = self.options.exists
            if self.options.limit:
                old_paths = paths
                paths = [p for p in paths
                         if p.cache.meta.size is None or  # if we have no known size don't limit it
                         search_exists or
                         not p.exists() and
                         p.cache.meta.size.mb < self.options.limit or
                         p.exists() and p.size != p.cache.meta.size and
                         (not log.info(f'Truncated transfer detected for {p}\n'
                                       f'{p.size} != {p.cache.meta.size}'))
                         and p.cache.meta.size.mb < self.options.limit]

                n_skipped = len(set(p for p in old_paths
                                    if p.is_broken_symlink()) - set(paths))

            if self.options.pretend:
                self._print_paths(paths)
                print(f'skipped = {n_skipped:<10}rate = {self.options.rate}')
                return

            if self.options.verbose:
                for p in paths:
                    print(p.cache.meta.as_pretty(pathobject=p))

            if self.options.fetch or self.options.refresh:
                from pyontutils.utils import Async, deferred
                hz = self.options.rate  # was 30
                limit = self.options.limit
                fetch = self.options.fetch
                if self.options.refresh:
                    Async(rate=hz)(deferred(path.remote.refresh)(
                        update_cache=True, update_data=fetch, size_limit_mb=limit)
                                   for path in paths)
                elif fetch:
                    def wrap(path):
                        def inner(*args, **kwargs):
                            try:
                                return path.cache.fetch(*args, **kwargs)
                            except Exception as e:
                                log.exception(e)
                                msg = f'Problem with path!\n{path}'
                                log.critical(msg)
                                raise e

                        return inner

                    Async(rate=hz)(deferred(wrap(path))(
                        size_limit_mb=limit)
                                   for path in paths)

            else:
                self._print_paths(paths)
                print(f'skipped = {n_skipped:<10}rate = {self.options.rate}')

    def feedback(self):
        file = self.options.feedback_file
        feedback = ' '.join(self.options.feedback)
        path = Path(file).resolve()
        eff = self.Integrator(path)
        # TODO pagenote and/or database
        print(eff, feedback)

    def missing(self):
        for path in self._paths:
            for rc in path.rchildren:
                if rc.is_broken_symlink():
                    m = rc.cache.meta
                    if m.file_id is None:
                        #print(rc)
                        print(m.as_pretty(pathobject=rc))
        #self.bfl.find_missing_meta()

    def xattrs(self):
        raise NotImplementedError('This used to populate the metastore, '
                                  'but currently does nothing.')

    def meta(self):
        if self.options.browser:
            import webbrowser

        self._cache_class._local_class = Path  # since we skipped _setup
        Path._cache_class = self._cache_class
        paths = self.paths
        if not paths:
            paths = self.cwd,

        old_level = log.level
        if not self.options.verbose:
            log.setLevel('ERROR')
        def inner(path):
            if self.options.uri or self.options.browser:
                if self.options.human or self.options.browser:
                    uri = path.cache.uri_human
                else:
                    uri = path.cache.uri_api

                print('+' + '-' * (len(uri) + 2) + '+')
                print(f'| {uri} |')
                if self.options.browser:
                    webbrowser.open(uri)

            try:
                cmeta = path.cache.meta
                if cmeta is not None:
                    if self.options.diff:
                        if path.is_dir():
                            print('It is not meaningful to diff directory metadata.')
                            return
                        if cmeta.checksum is None:
                            if not path.cache.local_object_cache_path.exists():
                                # we are going to have to go to network
                                self._setup()

                            cmeta.checksum = path.cache.checksum()  # FIXME :/
                        lmeta = path.meta
                        # id and file_id are fake in this instance
                        setattr(lmeta, 'id', None)
                        setattr(lmeta, 'file_id', None)
                        print(lmeta.as_pretty_diff(cmeta, pathobject=path,
                                                   human=self.options.human))
                    else:
                        print(cmeta.as_pretty(pathobject=path,
                                              human=self.options.human))

                    if self.options.context:
                        pc = path.cache
                        if not pc.is_organization():
                            org = pc.organization.local
                            if not pc.is_dataset():
                                ds = pc.dataset.local
                                inner(ds)
                            elif self.options.verbose:
                                inner(org)

                        print()

            except exc.NoCachedMetadataError:
                print(f'No metadata for {path}. Run `spc refresh {path}`')

        for path in paths:
            inner(path)

        log.setLevel(old_level)

    def goto(self):
        # TODO this needs an inverted index
        if self.options.remote_id.startswith('N:dataset:'):
            gen = self.cwd.children
        else:
            gen = self.cwd.rchildren

        for rc in gen:
            try:
                if rc.cache.id == self.options.remote_id:
                    if rc.is_broken_symlink() or rc.is_file():
                        rc = rc.parent

                    print(rc.relative_to(self.cwd).as_posix())
                    return
            except AttributeError as e:
                if not rc.skip_cache:
                    log.critical(rc)
                    log.error(e)

    def status(self):
        project_path = self.cwd.find_cache_root()
        if project_path is None:
            print(f'{self.cwd} is not in a project!')
            sys.exit(111)

        existing_files = [f for f in project_path.rchildren if f.is_file()]
        different = []
        for f in existing_files:
            try:
                cmeta = f.cache.meta
            except AttributeError:
                if f.skip_cache:
                    continue

            lmeta = f.meta if cmeta.checksum else f.meta_no_checksum
            # id and file_id are fake in this instance
            setattr(lmeta, 'id', None)
            setattr(lmeta, 'file_id', None)
            if lmeta.content_different(cmeta):
                if self.options.status:
                    print(lmeta.as_pretty_diff(cmeta, pathobject=f,
                                               human=self.options.human))
                else:
                    yield f, lmeta, cmeta

    def server(self):
        from sparcur.server import make_app
        if self.options.raw:
            # FIXME ...
            self.dataset_index = {d.meta.id:self.Integrator(d)
                                  for d in self.datasets}
        else:
            #from sparcur import export as ex
            #data = self._export(ex.Export).latest_ir
            #self.dataset_index = {d['id']:d for d in data['datasets']}
            pass

        report = Report(self)
        app, *_ = make_app(report)
        self.app = app  # debug only
        app.debug = False
        app.run(host='localhost', port=self.options.port, threaded=True)

    def stash(self, paths=None, stashmetafunc=lambda v:v):
        if paths is None:
            paths = self.paths

        stash_base = self.anchor.local.parent / 'stash'  # TODO move to operations
        to_stash = sorted(set(parent for p in paths for parent in
                               chain((p,), p.relative_to(self.anchor).parents)),
                           key=lambda p: len(p.as_posix()))

        if self.options.restore:
            # horribly inefficient, maybe build on a default dict?
            rcs = sorted((c for c in stash_base.rchildren if not c.is_dir()),
                         key=lambda p:p.as_posix(), reverse=True)
            for path in paths:
                relpath = path.relative_to(self.anchor)
                for p in rcs:
                    if p.parts[-len(relpath.parts):] == relpath.parts:
                        p.copy_to(path, force=True, copy_cache_meta=True)  # FIXME old remote may have been deleted, worth a check?
                        # TODO checksum? sync?
                        break

            breakpoint()

        else:
            stash = StashPath(stash_base, self._folder_timestamp)
            new_anchor = stash / self.anchor.name
            new_anchor.mkdir(parents=True)
            new_anchor.cache_init(self.anchor.meta, anchor=True)
            for p in to_stash[1:]:
                p = p.relative_to(self.anchor) if p.root == '/' else p
                new_path = new_anchor / p
                log.debug(p)
                log.debug(new_path)
                p = self.anchor.local / p
                if p.is_dir():
                    new_path.mkdir()
                    new_path.cache_init(p.cache.meta)
                else:
                    # TODO search existing stashes to see if
                    # we already have a stash of the file
                    log.debug(f'{p!r} {new_path!r}')
                    new_path.copy_from(p)
                    pc, npc = p.checksum(), new_path.checksum()
                    # TODO a better way to do this might be to
                    # treat the stash as another local for which
                    # the current local is the remote
                    # however this might require layering
                    # remote and remote remote metadata ...
                    assert pc == npc, f'\n{pc}\n{npc}'
                    cmeta = p.cache.meta
                    nmeta = stashmetafunc(cmeta)
                    log.debug(nmeta)
                    new_path.cache_init(nmeta)

            nall = list(new_anchor.rchildren)
            return nall

    def apinat(self):
        from pyontutils.core import populateFromJsonLd, OntGraph
        path_in = Path(self.options.path_in)
        path_out = Path(self.options.path_out)
        g = populateFromJsonLd(OntGraph(), path_in).write(path=path_out)

    def rmeta(self, use_cache_path=False, exist_ok=False):
        from pyontutils.utils import Async, deferred
        from sparcur.backends import PennsieveDatasetData as RemoteDatasetData
        dsr = self.datasets if use_cache_path else self.datasets_remote
        all_ = [RemoteDatasetData(r) for r in dsr]
        prepared = [bdd for bdd in all_ if not (exist_ok and bdd.cache_path.exists())]
        hz = self.options.rate
        if not self.options.debug:
            blobs = Async(rate=hz)(deferred(d)() for d in prepared)
        else:
            blobs = [d() for d in  prepared]

    def make_url(self):
        lu = {d.cache.id:d.cache for d in self.datasets_local}
        for iop in self.options.id_or_path:
            if iop.startswith('N:dataset:'):
                cache = lu[iop]
            else:
                cache = Path(iop).cache

            print(hfn.atag(cache.uri_human, cache.id))

    def show(self):
        if self.options.export:
            from sparcur import export as ex
            export = self._export(ex.Export, org_id=self.options.project_id)
            if self.options.json:
                export.latest_export_path.xopen(self.options.open)
            elif self.options.ttl:
                export.latest_ttl_path.xopen(self.options.open)

        elif self.options.schemas:
            print(self.options.export_schemas_path)
            self.options.export_schemas_path.xopen(self.options.open)  # NOTE it is a folder

        elif self.options.rmeta:
            from sparcur import datasources as ds
            Path(ds.PennsieveDatasetData.cache_base).xopen(self.options.open)

        elif self.options.protcur:
            latest_path = self.options.export_protcur_base / 'LATEST'
            latest_partial_path = self.options.export_protcur_base / 'LATEST_PARTIAL'
            path = (latest_partial_path if self.options.partial else latest_path) / 'protcur'
            suffix = '.json' if self.options.json else '.ttl'
            path.with_suffix(suffix).xopen(self.options.open)

    def sheets(self):
        from pyontutils import sheets as ps
        #from sparcur import sheets as ss

        if self.options.cache:
            from sparcur.sheets import Sheet, Organs, Affiliations
            Sheet._do_cache = True
            Sheet._re_cache = True
            for sh in (Organs, Affiliations):
                s = sh()
                s.fetch()
            return

        # get sheets
        sheets = {(s.name, s.sheet_name):s for s in subclasses(ps.Sheet)
                  if not s.__name__.startswith('_')}
        by_sn = {sn:s for (n, sn), s in sheets.items()}
        sn = self.options.sheet_name
        Sheet = by_sn[sn]
        Sheet.fetch_grid = False

        if self.options.update:
            sheet = Sheet(readonly=False)
            data = self._data_ir(self.options.project_id)
            # check that the ir is sane
            self._check_duplicates(data['datasets'])
            sheet.update_from_ir(data)
        else:
            sheet = Sheet()

        if sheet.sheet_name == 'Organs':
            h = sheet.row_object(0)
            #tech_hcells = [hcell for hcell in h.cells if 'technique' in hcell.value]
            #tech_cols = [hcell.column for hcell in tech_hcells]
            tech_cells = [
                cell
                for hcell in h.cells if 'technique' in hcell.value
                for cell in hcell.column.cells[1:] if cell.value]
            #sheet.fetch_grid = True
            sheet.fetch()

        if self.options.debug:
            breakpoint()

        return sheet

    ### sub dispatchers

    def report(self):
        report = Report(self)
        return report('report')

    def shell(self):
        """ drop into an shell with classes loaded """
        shell = Shell(self)
        return shell('shell')

    def fix(self):
        fix = Fix(self)
        return fix('fix')

    def fab(self):
        fix = Fab(self)
        return fix('fix')


class Report(reports.Report, Dispatcher):

    paths = Main.paths
    _paths = Main._paths

    _print_table = Main._print_table

    _pyru_loaded = False
    _data_ir = Main._data_ir  # done statically for server
    _protcur = Main._protcur  # done statically for server


class Shell(Dispatcher):
    # property ports
    paths = Main.paths
    _paths = Main._paths
    _build_paths = Main._build_paths
    datasets = Main.datasets
    datasets_local = Main.datasets_local
    stash = Main.stash

    def exit(self):
        """ useful for profiling startup time issues """
        self.options.project_id
        print('Peace.')
        return

    def sheets(self):
        from sparcur import sheets
        wev = sheets.WorkingExecVerb()
        wev.condense()
        organs = sheets.Organs(readonly=True)
        breakpoint()

    def default(self):
        datasets = list(self.datasets)
        datas = [self.Integrator(d).datasetdata for d in datasets]
        datasets_local = list(self.datasets_local)
        dsd = {d.meta.id:d for d in datasets}
        ds = datasets
        summary = self.summary
        org = self.Integrator(self.project_path)

        #asdf = []
        #for d in datasets:
            #asdf.append(list(d.children))

        breakpoint()

    def dates(self):
        paths = self.paths
        paths = paths if paths else list(self.datasets_local)
        asdf = {p:p.updated_cache_transitive() for p in paths}
        s = sorted(asdf.items(),
                   key=lambda kv: (1, kv[0]) if kv[-1] is None else (0, kv[-1]),
                   reverse=True)
        breakpoint()

    def more(self):
        p, *rest = self._paths
        if p.cache.is_dataset():
            intr = self.Integrator(p)
            j = JT(intr.data())
            iddf = intr.datasetdata.dataset_description.object  # finally
            #triples = list(f.triples)

        try:
            latest_datasets = self._export(None).latest_export['datasets']
        except:
            pass

        datasets = self.datasets
        datas = [self.Integrator(d).datasetdata for d in datasets]
        dsd = {d.meta.id:d for d in datasets}

        rcs = list(datasets[-1].rchildren)
        asdf = rcs[-1]
        urg = list(asdf.data)
        resp = asdf.data_headers

        did = self.cwd.cache.id
        r = self.cwd.remote
        if did in dsd:
            dsi = datas.index(dsd[did])
        else:
            dsi = 2

        d = datas[dsi]
        try:
            smf = d.submission.object
            smd = smf.data

            ddf = d.dataset_description.object
            ddd = ddf.data

            suf = d.subjects.object
            sud = suf.data

            saf = d.samples.object
            sad = saf.data

        except AttributeError as e:
            logd.error(f'{d} is missing some file')

        breakpoint()

    def affil(self):
        from pyontutils.utils import Async, deferred
        from sparcur import sheets
        a = sheets.Affiliations()
        m = a.mapping
        rors = sorted(set(_ for _ in m.values() if _))
        #dat = Async(rate=5)(deferred(lambda r:r.data)(i) for i in rors)
        dat = [r.data for r in rors]  # once the cache has been populated
        breakpoint()

    def protocols(self):
        """ test protocol identifier functionality """
        org = self.Integrator(self.project_path)
        from pyontutils.utils import Async, deferred
        skip = '"none"', 'NA', 'no protocols', 'take protocol from other spreadsheet, ', 'na'
        asdf = [us for us in sorted(org.organs_sheet.byCol.protocol_url_1)
                if us not in skip and us and ',' not in us]
        wat = [idlib.Auto(_) for _ in asdf]
        inst = [i for i in wat]
        res = Async(rate=5)(deferred(i.dereference)(idlib.Auto) for i in inst)
        pis = [i for i in res if isinstance(i, idlib.Pio)]
        #dat = Async(rate=5)(deferred(lambda p: p.data)(i) for i in pis)
        #dois = [d['protocol']['doi'] for d in dat if d]
        dois = [p.doi for p in pis]
        breakpoint()

    def integration(self):
        from protcur.analysis import protc, Hybrid
        #from sparcur.sheets import Organs, Progress, Grants, ISAN, Participants, Protocols as ProtocolsSheet
        from sparcur.protocols import ProtcurData
        p, *rest = self._paths
        intr = self.Integrator(p)
        j = JT(intr.data)
        pj = list(intr.protocol_jsons)
        pc = list(intr.triples_exporter.protcur)
        #apj = [pj for c in intr.anchor.children for pj in c.protocol_jsons]
        breakpoint()


class Fab(Dispatcher):

    paths = Main.paths
    _paths = Main._paths

    _print_table = Main._print_table

    def meta(self):
        """ """
        # XXX FIXME this is a temporary hack until we fix the ability
        # to run export
        pass


class Fix(Shell):

    def default(self):
        pass

    def bf_to_pn(self):
        # TODO placeholder for fixes needed for backend transition
        pass

    def cache(self):
        new_dirs = set()
        def make_target(path):
            pm = path._xattr_meta
            id = PennsieveId(pm.id, file_id=pm.file_id)
            uuid = id.uuid
            # XXX see PennsieveCache.cache_key for the pattern
            new_dir = uuid[:2]
            new_dirs.add(lod / new_dir)
            target_suffix = f'{new_dir}/{uuid}-{id.file_id}'
            target = lod / target_suffix
            return target

        lod = self.cwd.cache.local_objects_dir
        to_move = [c for c in lod.children if c.is_file()]
        types = set(p.name.split(':')[1] for p in to_move)
        assert types == {'package'}  # there should only be packages in here
        st = [(source, make_target(source)) for source in to_move]
        to_mkdir = [d for d in new_dirs if not d.exists()]
        _ = [d.mkdir() for d in to_mkdir]
        _ = [s.rename(t) for s, t in st]

    def mismatch(self):
        """ once upon a time it was (still at the time of writing) possible
            to update meta on an existing file without preserving the old data
            AND without downloading the new data (insanity) this can be used to fix
            those cases, preserving the old version """
        oops = list(self.parent.status())
        [print(lmeta.as_pretty_diff(cmeta, pathobject=path, human=self.options.human))
         for path, lmeta, cmeta in oops]
        paths = [p for p, *_ in oops]

        def sf(cmeta):
            nmeta = aug.PathMeta(id=cmeta.old_id)
            assert nmeta.id, f'No old_id for {cmeta}'
            return nmeta

        nall = self.stash(paths, stashmetafunc=sf)
        [print(n.cache.meta.as_pretty(n)) for n in nall]
        breakpoint()
        # once everything is in order and backed up 
        # [p.cache.fetch() for p in paths]

    def duplicates(self):
        all_ = defaultdict(list)
        if not self.options.path:
            paths = self.anchor.local.rchildren
        else:
            paths = self.paths

        for rc in paths:
            if rc.cache is None:
                if not rc.skip_cache:
                    log.critical(f'WHAT THE WHAT {rc}')

                continue

            all_[rc.cache.id].append(rc)

        def mkey(p):
            mns = p.cache.meta
            return (not bool(mns),
                    not bool(mns.updated),
                    -mns.updated.timestamp())

        dupes = {i:sorted(l, key=mkey)#, reverse=True)
                 for i, l in all_.items() if len(l) > 1}
        dv = list(dupes.values())
        deduped = [a.dedupe(b, pretend=True) for a, b, *c in dv
                   if (not log.warning(c) if c else not c)
        ]  # FIXME assumes a single dupe...
        to_remove = [d for paths, new in zip(dv, deduped)
                     for d in paths if d != new]
        to_remove_size = [p for p in to_remove if p.cache.meta.size is not None]
        #[p.unlink() for p in to_remove if p.cache.meta.size is None] 
        breakpoint()


def main():
    time_now = GetTimeNow()
    from docopt import docopt, parse_defaults
    args = docopt(__doc__, version='spc 0.0.0')
    defaults = {o.name:o.value if o.argcount else None for o in parse_defaults(__doc__)}

    logpath = Path(args['--log-path'])
    if not logpath.exists():
        logpath.mkdir(parents=True)

    try:
        logfile = logpath / time_now.START_TIMESTAMP_SAFE  # FIXME configure and switch
        if False:  # python file loggers are ... not capable of what we need it seems
            bind_file_handler(logfile)

        options = Options(args, defaults)
        main = Main(options, time_now)
        if main.options.debug:
            print(main.options)

        try:
            main()
        except BaseException as e:
            log.exception(e)
            print()
            cwd = main.cwd if hasattr(main, 'cwd') else Path.cwd()
            msg = f'Command failed!\n{cwd}\n{options!r}'
            raise exc.SparCurError(msg) from e

    finally:
        if logfile.exists() and logfile.size == 0:
            # close log handlers to safely unlink on windows
            [h.close() for h in log.handlers]
            logfile.unlink()

    if options.profile:
        exit = time()
        a = start_middle - start
        b = (stop - start_middle)
        c = (stop - start)
        d = (exit - stop)
        e = (exit - start)
        s = f'{a} + {b} = {c}\n{c} + {d} = {e}'
        print(s)


if __name__ == '__main__':
    main()



import unittest
from pathlib import Path
import pytest
from sparcur import pipelines as pipes
from .common import (examples_root,
                     project_path,
                     RDHPN,
                     )


class TestDatasetDescription(unittest.TestCase):
    source = examples_root / 'dd-pie.csv'

    def test_dd_pie_p(self):
        p = pipes.DatasetDescriptionFilePipeline(self.source, None, None)
        data = p.data
        # TODO test as subpipeline ?


class PipelineHelper:

    @classmethod
    def setUpClass(cls):
        cls.project_path = project_path
        cls.datasets = list(cls.project_path.children)

    def _path_to_pipe(self, dataset_path):
        """ FIXME TODO this needs to be simplified """
        class context:
            path = dataset_path.resolve()
            id = path.id
            uri_api = path.as_uri()
            uri_human = path.as_uri()

        class lifters:
            # minimal set
            id = context.id
            remote = context.path._cache_class._remote_class._remote_type
            folder_name = context.path.name
            uri_api = context.uri_api
            uri_human = context.uri_human
            timestamp_export_start = None

            # extended requirements (annoying)
            # FIXME these need to be removed
            techniques = 'FAKE TECHNIQUE'
            award_manual = 'FAKE TOTALLY NOT AN AWARD'
            modality = 'THE MODALITY THE HAS BECOME ONE WITH NOTHINGNESS'
            organ_term = 'ilxtr:NOGGIN'  # expects a curie or iri
            protocol_uris = ('https://example.org/TOTALLY-NOT-A-REAL-URI',)
            affiliations = lambda _: None

        pipe = pipes.PipelineEnd(dataset_path, lifters, context)
        return pipe

    def test_pipeline_end(self):
        pipelines = []
        for dataset_path in self.test_datasets:
            pipe = self._path_to_pipe(dataset_path)
            pipelines.append(pipe)

        bads = []
        fails = []
        errors = []
        for p in pipelines:
            try:
                d = p.data
                if 'errors' in d:
                    errors.append(d.pop('errors'))
                    fails.append(d)
                    if 'submission_errors' in d['status']:
                        d['status'].pop('submission_errors')
                    if 'curation_errors' in d['status']:
                        d['status'].pop('curation_errors')
                    if 'errors' in d['inputs']:
                        d['inputs'].pop('errors')
            except Exception as e:
                raise e
                bads.append((e, p))

        assert not bads, bads


class TestPipelines(unittest.TestCase):
    pass


class TestPipelinesRealPN(RDHPN, PipelineHelper, unittest.TestCase):
    # RealDataHelper needs to resolve first to get correct setUpClass
    pass



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::clone.py][clone.py]]
from pathlib import Path


# clone top level
def from_path_id_and_backend_project_top_level(parent_path,
                                               project_id,
                                               RemotePath,
                                               symlink_objects_to=None,):
    """ given the enclosing path to clone to, the project_id, and a fully
        configured (with Local and Cache) backend remote path, anchor the
        project pointed to by project_id along with the first level of children """

    project_path = _from_path_id_remote_project(parent_path,
                                                project_id,
                                                RemotePath,
                                                symlink_objects_to)
    return _from_project_path_top_level(project_path)


def from_path_project_backend_id_dataset(parent_path,
                                         project_id,
                                         dataset_id,
                                         RemotePath,
                                         symlink_objects_to=None,):
    project_path = _from_path_id_remote_project(parent_path,
                                                project_id,
                                                RemotePath,
                                                symlink_objects_to)
    return _from_project_path_id_dataset(project_path, dataset_id)


def _from_path_id_remote_project(parent_path, project_id, RemotePath, symlink_objects_to):
    RemotePath.init(project_id)  # calling init is required to bind RemotePath._api
    anchor = RemotePath.smartAnchor(parent_path)
    anchor.local_data_dir_init(symlink_objects_to=symlink_objects_to)
    project_path = anchor.local
    return project_path


def _from_project_path_top_level(project_path):
    """ given a project path with existing cached metadata
        pull the top level children

        WARNING: be VERY careful about using this because it
        does not gurantee that rmeta is available to mark
        sparse datasets. It may be the case that the process
        will fail if the rmeta is missing, or it may not. Until
        we are clear on the behavior this warning will stay
        in place. """
    # this is a separate function in case the previous step fails
    # which is also why it is hidden, it makes too many assuptions
    # to be used by itself

    anchor = project_path.cache
    list(anchor.children)  # this fetchs data from the remote path to the local path
    return project_path  # returned instead of anchor & children because it is needed by next phase


def _from_project_path_id_dataset(project_path, dataset_id):
    anchor = project_path.cache
    remote = anchor._remote_class(dataset_id)
    cache = anchor / remote
    return cache.local


def main(parent_path=None,
         project_id=None,
         parent_parent_path=Path.cwd(),
         project_id_auth_var='remote-organization',  # FIXME move default to clifun
         symlink_objects_to=None,
         id=None,
         dataset_id=None,
         **kwargs):
    """ clone a project into a random subfolder of the current folder
        or specify the parent path to clone into """

    from sparcur.simple.utils import backend_pennsieve

    if parent_path is None:
        breakpoint()  # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX FIXME
        import tempfile
        parent_path = Path(tempfile.mkdtemp(dir=parent_parent_path))

    if project_id is None:
        from sparcur.config import auth
        from sparcur.utils import PennsieveId
        project_id = auth.get(project_id_auth_var)
        project_id = PennsieveId(project_id)  # FIXME abstract the hardcoded backend

    RemotePath = backend_pennsieve()

    if id and dataset_id:
        # FIXME doesn't check for existing so if the name changes we get duped folders
        # this issue possibly upstream in retrieve, clone just clones whatever you tell
        # it to clone, but maybe it should check the existing metadata and fail or warn?
        dataset_path = from_path_project_backend_id_dataset(
            parent_path,
            project_id,
            id,  # FIXME multiple datasets
            RemotePath,
            symlink_objects_to,)

        return dataset_path

    project_path = from_path_id_and_backend_project_top_level(
        parent_path,
        project_id,
        RemotePath,
        symlink_objects_to,)

    return project_path


if __name__ == '__main__':
    from sparcur.simple.utils import pipe_main
    pipe_main(main, after=print)
# clone.py ends here



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::dataset-extract][dataset-extract]]
from sparcur import datasets as dat
from sparcur import pipelines as pipes
from sparcur import exceptions as exc
from sparcur.utils import log, logd
from sparcur.paths import Path
from sparcur.backends import PennsieveDatasetData
from sparcur.simple.utils import combinate, multiple, early_failure, DataWrapper
from sparcur.simple.fetch_metadata_files import fetch_prefixes


class ManifestFiles(DataWrapper):
    """ wrapper for manifest files. """


def merge_manifests(vs):
    """ Merge individual manifest records into the same list """
    # FIXME massive hack :/
    k = 'manifest_file'
    # FIXME errors key ... ? is it allowed up there? it shouldn't be ...
    # FIXME {'content': m}
    return ManifestFiles([m for v in vs for m in v.data[k]])


def object_from_find_path(glob_prefix, object_from_path_function, glob_type='glob', onfail=None):
    """ Return a function that will find files that start with glob_prefix"""
    # FIXME should be in utils but depends on fetch_prefixes
    if glob_prefix not in dict(fetch_prefixes):
        raise ValueError('glob_prefix not in fetch_prefixes! '
                         f'{glob_prefix!r} not in {fetch_prefixes}')
    def func(path, *args, **kwargs):
        ds = dat.DatasetStructure(path)
        rpath = None
        for rpath in ds._abstracted_paths(glob_prefix, sandbox=True):
            yield object_from_path_function(rpath, *args, **kwargs)

        if rpath is None and onfail is not None:
            raise onfail(f'No match for {glob_prefix} in {path.name}')

    return func

# file type -> dataset blob key indirection

_TOP = object()  # SIGH SIGH SIGH always need a escape hatch

otkm = {ThingFilePath.obj:prefix + '_file' for prefix, ThingFilePath
        in dat.DatasetStructure.sections.items()}
otkm[ManifestFiles] = 'manifest_file'
otkm[PennsieveDatasetData] = 'remote_dataset_metadata'
otkm[type(dat.DatasetStructure())] = 'structure'  # hack around Pathlib type mangling
otkm[type(dat.DatasetMetadata())] = _TOP

# stream of objects -> place in dataset blob

def dataset_raw(*objects, object_type_key_map=otkm):
    data = {}
    log.debug(objects)
    #path_structure, description, subjects, samples, submission, manifests, *rest = objects
    for obj in objects:
        log.debug(obj)
        key = object_type_key_map[type(obj)]
        try:
            if key is not _TOP:
                data.update({key: obj.data})
            else:
                data.update(obj.data)
        except Exception as e:
            # FIXME current things that leak through
            # MalformedHeaderError
            # something in the objects list is a dict
            breakpoint()
            pass

    return data


# path + version -> python object

# TODO how to attach and validate schemas orthogonally in this setting?
# e.g. so that we can write dataset_1_0_0 dataset_1_2_3 etc.

# we capture version as early as possible in the process, yes we
# could also gather all the files and folders and then pass the version
# in as an argument when we validate their structure, but there are
# elements of the locations or names of those files that might depend
# on the template version, therefore we get maximum flexibility by only
# need to look for the dataset description file
def description(path):          return dat.DatasetDescriptionFilePath(path).object
def submission(path, version):  return dat.SubmissionFilePath(path).object_new(version)
def subjects(path, version):    return dat.SubjectsFilePath(path).object_new(version)
def samples(path, version):     return dat.SamplesFilePath(path).object_new(version)
def manifest(path, version):    return dat.ManifestFilePath(path).object_new(version)

# dataset path -> python object

def from_path_remote_metadata(path): return PennsieveDatasetData(path.cache)
def from_path_local_metadata(path): return dat.DatasetMetadata(path)
from_path_dataset_description = object_from_find_path('dataset_description', description,
                                                      onfail=exc.MissingFileError)

comb_metadata = combinate(
    # dataset description is not included here because it is special
    # see from_path_dataset_raw for details
    from_path_remote_metadata,
    from_path_local_metadata,
)

# dataset path + version -> python object

def from_path_dataset_path_structure(path, version): return dat.DatasetStructure(path)
from_path_subjects   = object_from_find_path('subjects',            subjects)
from_path_samples    = object_from_find_path('samples',             samples)
from_path_submission = object_from_find_path('submission',          submission)
from_path_manifests  = multiple(object_from_find_path('manifest',   manifest,
                                                      'rglob'),
                                merge_manifests)

# combinate all the individual dataset path + version -> data functions

comb_dataset = combinate(
    #from_path_dataset_description,  # must be run prior to combination to get version
    from_path_dataset_path_structure,
    from_path_subjects,
    from_path_samples,
    from_path_submission,
    from_path_manifests,
    #from_path_file_metadata,  # this must wait until 2nd fetch phase
    )

# dataset path -> raw data

def from_path_dataset_raw(dataset_path):
    """ Main entry point for getting dataset metadata from a path. """
    gen  = from_path_dataset_description(dataset_path)
    try:
        ddo = dataset_description_object = next(gen)
    except exc.MissingFileError as e:
        # TODO return a stub with embedded error
        logd.critical(e)
        dataset_blob = dataset_raw(*comb_metadata(dataset_path))
        return early_failure(dataset_path, e, dataset_blob)

    try:
       next(gen)
       # TODO return a stub with embedded error
    except StopIteration:
        pass

    data = ddo.data
    ddod = type('ddod', tuple(), {'data': data})
    dtsv = data['template_schema_version']
    return dataset_raw(ddo, *comb_metadata(dataset_path), *comb_dataset(dataset_path, dtsv))


# unused

def from_path_file_metadata(path, _version):  # FIXME doesn't go in this file probably
    # FIXME this is going to depend on the manifests
    # and a second fetch step where we kind of cheat
    # and prefetch file files we know we will need
    pass


def from_export_path_protocols_io_data(curation_export_json_path): pass
def protocols_io_ids(datasets): pass
def protocols_io_data(protocols_io_ids): pass

def from_group_name_protcur(group_name): pass
def protcur_output(): pass

def summary(datasets, protocols_io_data, protcur_output): pass

def from_path_summary(project_path):
    dataset_path_structure
    summary((
        dataset(
            dataset_path_structure,
            dataset_description,
            subjects,
            samples,
            submission,
            manifests,
            *rest
)))


def main(path=Path.cwd(), id=None, dataset_id=tuple(), time_now=None,
         export_path=None, export_parent_path=None, _entry_point=False, **kwargs):
    # TODO path from dataset_id and retrieve conventions? or just pass path from retrieve final output?
    # TODO parallelize if multiple paths
    # This assumes that all retrieve operations have
    # finished successfully for the current dataset
    from sparcur.simple.export import prepare_dataset_export, export_blob
    if id is not None:  # XXX note that id should be dataset_id # TODO parent_path ??
        uuid = id.uuid  # FIXME hardcoded backend assumption
        path = path / uuid / 'dataset'  # FIXME hardcoded see invariant_path
        path = path.resolve()  # make sure that invariant_path is expanded as we expect.

    cache = path.cache
    if not cache.is_dataset():
        raise TypeError('can only run on a single dataset')

    if _entry_point:
        if export_parent_path is None:
            export_parent_path = prepare_dataset_export(path, export_path)

        kwargs.update({'export_path': export_path,
                       'export_parent_path': export_parent_path,
                       'time_now': time_now,})

    dataset_raw = from_path_dataset_raw(path)

    if _entry_point:
        export_blob_path = export_blob(dataset_raw, 'ir.json', **kwargs)
        return export_blob_path
    else:
        return dataset_raw


if __name__ == '__main__':
    import pprint
    from sparcur.simple.utils import pipe_main
    pipe_main(main, after=pprint.pprint)
# dataset-extract ends here



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::*_*Test*_][_*Test*_:1]]
from sparcur.simple.utils import pipe_main

def test_pipe_main():
    def main(id=None, project_path=None, **kwargs):
        print(id, project_path, kwargs)

    pipe_main(main, argv=['sparcur-simple'])
# _*Test*_:1 ends here
